"""
Put all functions defined for gdb in here.

This file might be changed into a module in the future.
"""

from __future__ import annotations

import functools
from collections.abc import Callable
from typing import Any

import gdb

import pwndbg.aglib.argv
import pwndbg.aglib.elf
import pwndbg.aglib.proc
import pwndbg.aglib.vmmap
import pwndbg.dbg_mod
from pwndbg.lib.common import hex2ptr_common

functions: list[_GdbFunction] = []


def GdbFunction(only_when_running: bool = False) -> Callable[..., Any]:
    return functools.partial(_GdbFunction, only_when_running=only_when_running)


class _GdbFunction(gdb.Function):
    def __init__(self, func: Callable[..., Any], only_when_running: bool) -> None:
        self.name = func.__name__
        self.func = func
        self.only_when_running = only_when_running
        self.__doc__ = func.__doc__.strip()

        assert func.__doc__ and "The function must have a docstring."
        _first_line = self.__doc__.split("\n")[0]
        assert len(_first_line) <= 80 and (
            "The first line of the function's docstring should be short,"
            " as it is printed with `help function`."
        )
        assert _first_line[-1] == "." and (
            "The first line should be a standalone sentence, as it is "
            "printed alone with `help function`."
        )
        assert (
            "Example:\n" in func.__doc__
            and "Convenience functions need to provide a usage example."
        )

        functions.append(self)

        super().__init__(self.name)

        functools.update_wrapper(self, func)

    def invoke(self, *args: gdb.Value) -> Any:
        if self.only_when_running and not pwndbg.aglib.proc.alive():
            # Returning empty string is a workaround that we can't stop e.g. `break *$rebase(offset)`
            # Thx to that, gdb will print out 'evaluation of this expression requires the target program to be active'
            return ""

        return self.func(*args)

    def __call__(self, *args: gdb.Value) -> Any:
        return self.invoke(*args)


@GdbFunction(only_when_running=True)
def rebase(addr: gdb.Value | int) -> int:
    """
    Return address rebased onto the executable's mappings.

    Example:
    ```
    pwndbg> p/x $rebase(0xd9020)
    $1 = 0x55555562d020
    pwndbg> vmmap
    0x555555554000     0x55555556f000 r--p    1b000      0 /usr/bin/bash
    0x55555556f000     0x55555562d000 r-xp    be000  1b000 /usr/bin/bash
    0x55555562d000     0x55555565e000 r--p    31000  d9000 /usr/bin/bash
    [...]
    pwndbg> p $rebase(0xd9020) == 0x555555554000 + 0xd9020
    $2 = 1
    pwndbg> tele $rebase(0xd9020)
    00:0000│  0x55555562d020 ◂— 0x204900636f6c6c61 /* 'alloc' */
    01:0008│  0x55555562d028 ◂— 'have no name!'
    02:0010│  0x55555562d030 ◂— 0x65720021656d616e /* 'name!' */
    03:0018│  0x55555562d038 ◂— 'adline stdin'
    [...]
    ```
    """
    base = pwndbg.aglib.elf.exe().address
    return base + int(addr)


@GdbFunction(only_when_running=True)
def base(name_pattern: gdb.Value | str):
    """
    Return the base address of the first memory mapping containing the given name.

    Example:
    ```
    pwndbg> p/x $base("libc")
    $4 = 0x7ffff7d4b000
    pwndbg> vmmap libc
        0x7ffff7d4a000     0x7ffff7d4b000 rw-p     1000  6e000 /usr/lib/libncursesw.so.6.5
    ►   0x7ffff7d4b000     0x7ffff7d6f000 r--p    24000      0 /usr/lib/libc.so.6
    ►   0x7ffff7d6f000     0x7ffff7ed6000 r-xp   167000  24000 /usr/lib/libc.so.6
    ►   0x7ffff7ed6000     0x7ffff7f2b000 r--p    55000 18b000 /usr/lib/libc.so.6
    ►   0x7ffff7f2b000     0x7ffff7f2f000 r--p     4000 1e0000 /usr/lib/libc.so.6
    ►   0x7ffff7f2f000     0x7ffff7f31000 rw-p     2000 1e4000 /usr/lib/libc.so.6
        0x7ffff7f31000     0x7ffff7f39000 rw-p     8000      0 [anon_7ffff7f31]
    pwndbg> tele $base(\\"libc\\")+0x1337
    00:0000│  0x7ffff7d4c337 ◂— 0x80480a04214000f0
    01:0008│  0x7ffff7d4c33f ◂— 0x8040c02204452040
    02:0010│  0x7ffff7d4c347 ◂— 0x20042400000200
    03:0018│  0x7ffff7d4c34f ◂— 0x20 /* ' ' */
    [...]
    ```

    Beware of accidentally matching the wrong mapping. For instance, if the loaded
    executable contained the string "libc" anywhere in it's path, it would've been
    returned.
    """
    if isinstance(name_pattern, gdb.Value):
        name = name_pattern.string()
    else:
        name = name_pattern

    for p in pwndbg.aglib.vmmap.get():
        if name in p.objfile:
            return p.vaddr

    raise gdb.GdbError(f"$base error: No mapping named '{name}'")


@GdbFunction()
def hex2ptr(hex_string: gdb.Value | str) -> int:
    """
    Converts a hex string to a little-endian address and returns the address.

    Example:
    ```
    pwndbg> p/x $hex2ptr("20 74 ed f7 ff 7f")
    $1 = 0x7ffff7ed7420
    pwndbg> p/x $hex2ptr("2074edf7ff7f")
    $2 = 0x7ffff7ed7420
    pwndbg> distance '$base("libc")' '$hex2ptr("20 74 ed f7 ff 7f")'
    0x7ffff7d4b000->0x7ffff7ed7420 is 0x18c420 bytes (0x31884 words)
    ```

    Especially useful for quickly converting pwntools output.
    """
    if isinstance(hex_string, gdb.Value):
        hex_string = hex_string.string()

    hex_string = hex_string.replace(" ", "")
    pointer = hex2ptr_common(hex_string)
    return pointer


@GdbFunction(only_when_running=True)
def argc() -> int:
    """
    Get the number of program arguments.
    Evaluates to argc.

    Example:
    ```
    pwndbg> p $argc()
    $1 = 2
    pwndbg> argv
    00:0000│  0x7fffffffe288 —▸ 0x7fffffffe659 ◂— '/usr/bin/cat'
    01:0008│  0x7fffffffe290 —▸ 0x7fffffffe666 ◂— 'gdbinit.py'
    02:0010│  0x7fffffffe298 ◂— 0
    ```
    """
    return pwndbg.aglib.argv.argc()


@GdbFunction(only_when_running=True)
def argv(index: gdb.Value) -> gdb.Value:
    """
    Get the n-th program argument.
    Evaluate argv on the supplied value.

    Example:
    ```
    pwndbg> p $argv(0)
    $11 = (signed char *) 0x7fffffffe666 "/usr/bin/sh"
    pwndbg> argv
    00:0000│  0x7fffffffe2a8 —▸ 0x7fffffffe666 ◂— '/usr/bin/sh'
    01:0008│  0x7fffffffe2b0 ◂— 0
    ```
    """
    val = pwndbg.aglib.argv.argv(int(index))
    if val is None:
        raise gdb.GdbError("Arg not found")
    return dbg_value_to_gdb(val)


@GdbFunction(only_when_running=True)
def environ(env_name: gdb.Value) -> gdb.Value:
    """
    Get an environment variable by name.
    Evaluate getenv() on the supplied value.

    Example:
    ```
    pwndbg> p $environ("LANG")
    $2 = (signed char *) 0x7fffffffebfb "LANG=en_US.UTF-8"
    ```
    """
    name = env_name.string()
    if not name:
        raise gdb.GdbError("No environment variable name provided")

    val = pwndbg.aglib.argv.environ(name)
    if val is None:
        raise gdb.GdbError("Environ not found")
    return dbg_value_to_gdb(val)


@GdbFunction(only_when_running=True)
def envp(index: gdb.Value) -> gdb.Value:
    """
    Get the n-th environment variable.
    Evaluate envp on the supplied value.

    Example:
    ```
    pwndbg> p $envp(0x3F)
    $13 = (signed char *) 0x7fffffffef7d "LANG=en_US.UTF-8"
    pwndbg> p $envp(0x3F) == $environ("LANG")
    $14 = 1
    ```
    """
    val = pwndbg.aglib.argv.envp(int(index))
    if val is None:
        raise gdb.GdbError("Environ not found")
    return dbg_value_to_gdb(val)


def dbg_value_to_gdb(d: pwndbg.dbg_mod.Value) -> gdb.Value:
    from pwndbg.dbg_mod.gdb import GDBValue

    assert isinstance(d, GDBValue)
    return d.inner


@GdbFunction(only_when_running=True)
def fsbase(offset: gdb.Value = gdb.Value(0)) -> int:
    """
    Get the value of the FS segment register.
    Only valid on x86(-64).

    Example:
    ```
    pwndbg> p/x $fsbase()
    $3 = 0x7ffff7cdab80
    pwndbg> p $fs_base == $fsbase()
    $4 = 1
    pwndbg> x/gx $fsbase(0x28)
    0x7ffff7cdaba8:	0x4da926e1668e5a00
    pwndbg> x/gx $fsbase(0x30)
    0x7ffff7cdabb0:	0x190a86d93bccf0ad
    pwndbg> tls
    Thread Local Storage (TLS) base: 0x7ffff7cdab80
    TLS is located at:
        0x7ffff7cda000     0x7ffff7cdc000 rw-p     2000      0 [anon_7ffff7cda]
    Dumping the address:
    tcbhead_t @ 0x7ffff7cdab80
        0x00007ffff7cdab80 +0x0000 tcb                  : 0x7ffff7cdab80
        0x00007ffff7cdab88 +0x0008 dtv                  : 0x7ffff7cdb4f0
        0x00007ffff7cdab90 +0x0010 self                 : 0x7ffff7cdab80
        0x00007ffff7cdab98 +0x0018 multiple_threads     : 0x0
        0x00007ffff7cdab9c +0x001c gscope_flag          : 0x0
        0x00007ffff7cdaba0 +0x0020 sysinfo              : 0x0
        0x00007ffff7cdaba8 +0x0028 stack_guard          : 0x4da926e1668e5a00
        0x00007ffff7cdabb0 +0x0030 pointer_guard        : 0x190a86d93bccf0ad
        [...]
    pwndbg> canary
    [...]
    Canary    = 0x4da926e1668e5a00 (may be incorrect on != glibc)
    [...]
    ```
    FS will usually point to the start of the TLS. If you're not providing an
    offset, it is usually easier to use GDB's builtin $fs_base variable.
    """
    if pwndbg.aglib.arch.name not in ("i386", "x86-64"):
        raise gdb.GdbError("This function is only valid on i386 and x86-64.")

    return pwndbg.aglib.regs.fsbase + int(offset)


@GdbFunction(only_when_running=True)
def gsbase(offset: gdb.Value = gdb.Value(0)) -> int:
    """
    Get the value of the GS segment register.
    Only valid on x86(-64).

    Example:
    ```
    pwndbg> p/x $gsbase()
    $1 = 0x0
    ```
    The value of the GS register is more interesting when doing kernel debugging:
    ```
    pwndbg> p/x $gsbase()
    $1 = 0xffff999287a00000
    pwndbg> tele $gsbase()
    00:0000│  0xffff999287a00000 ◂— 0
    ... ↓     4 skipped
    05:0028│  0xffff999287a00028 ◂— 0xd6aa9b336d52a400
    06:0030│  0xffff999287a00030 ◂— 0
    07:0038│  0xffff999287a00038 ◂— 0
    pwndbg> p $gsbase() == $gs_base
    $2 = 1
    ```
    If you're not providing an offset, it is usually easier to use GDB's
    builtin $gs_base variable.
    """
    if pwndbg.aglib.arch.name not in ("i386", "x86-64"):
        raise gdb.GdbError("This function is only valid on i386 and x86-64.")
    return pwndbg.aglib.regs.gsbase + int(offset)



"""
The Pwndbg REPL that is the interface to all debugging on LLDB.

Pwndbg has an event system that allows it to react to events in the process
being debugged, such as when new executable modules get added to the its address
space, when the value of memory and registers change, and pretty much all
possible changes to its execution state. We'd like to have the event system work
the same way under LLDB as it does under GDB.

Fortunately for us, the events types that are native to LLDB map really well to
the event types in GDB and Pwndbg. Very, very unfortunately for us, however,
that's basically where our luck ends.

LLDB, as of version 18, only provides two ways to capture events: registering
directly with the broadcaster, or registering globally. The former is not
available to us in the standard LLDB REPL, as we don't get access to the process
object until after it's been launched[1]. Likewise for the latter, as the
interactive debugger will register to receive the global process state change
events before everyone else, and LLDB doesn't allow for multiple listeners for
the same event bits in the same event class[2].

This leaves us with handling process management ourselves as the only option we
really have to implement event dispatch in Pwndbg. Easy, right? We can just
hijack the commands that deal with target and process creation, and leave
everything else untouched. Unfortunately for us, again, shadowing builtin
commands is simply not allowed[3][4].

So, really, all that's left for us is either implement our own REPL, or get rid
of the event system.

[1]: https://discourse.llvm.org/t/understanding-debugger-launch-events-sequence/39717/2
[2]: https://github.com/llvm/llvm-project/blob/3b5b5c1ec4a3095ab096dd780e84d7ab81f3d7ff/lldb/source/Utility/Broadcaster.cpp#L409
[3]: https://github.com/llvm/llvm-project/blob/3b5b5c1ec4a3095ab096dd780e84d7ab81f3d7ff/lldb/source/Commands/CommandObjectCommands.cpp#L439
[4]: https://github.com/llvm/llvm-project/blob/3b5b5c1ec4a3095ab096dd780e84d7ab81f3d7ff/lldb/source/Interpreter/CommandInterpreter.cpp#L1157
"""

from __future__ import annotations

import argparse
import asyncio
import os
import re
import shutil
import signal
import sys
from asyncio import CancelledError
from collections.abc import Awaitable
from collections.abc import Callable
from collections.abc import Coroutine
from contextlib import contextmanager
from io import BytesIO
from io import TextIOWrapper
from typing import Any
from typing import BinaryIO

import lldb
from typing_extensions import override

import pwndbg
import pwndbg.dbg_mod
from pwndbg.color import message
from pwndbg.dbg_mod import EventType
from pwndbg.dbg_mod.lldb import LLDB
from pwndbg.dbg_mod.lldb import LLDBPythonState
from pwndbg.dbg_mod.lldb import OneShotAwaitable
from pwndbg.dbg_mod.lldb.pset import InvalidParse
from pwndbg.dbg_mod.lldb.pset import pget
from pwndbg.dbg_mod.lldb.pset import pset
from pwndbg.dbg_mod.lldb.repl.io import get_io_driver
from pwndbg.lib.tips import color_tip
from pwndbg.lib.tips import get_tip_of_the_day

HAS_FZF = shutil.which("fzf") is not None
if HAS_FZF:
    from pwndbg.dbg_mod.lldb.repl.fuzzy import PROMPT
    from pwndbg.dbg_mod.lldb.repl.fuzzy import get_prompt_session
    from pwndbg.dbg_mod.lldb.repl.fuzzy import wrap_with_history
else:
    from pwndbg.dbg_mod.lldb.repl.readline import PROMPT
    from pwndbg.dbg_mod.lldb.repl.readline import enable_readline
    from pwndbg.dbg_mod.lldb.repl.readline import wrap_with_history


class UserCancelledError(CancelledError):
    """
    Internal cancellation exception used by the LLDB CLI.

    Sometimes, it's necessary to cancel both commands and subroutines mid-execution. This is an internal exception type
    that is used in these conditions.
    """

    pass


def print_error(msg: str, *args):
    """
    Print an error message in the style of the LLDB CLI.
    """
    print(message.error("error:"), msg, *args)


def print_warn(msg: str, *args):
    """
    Print a warning message in the style of the LLDB CLI.
    """
    print(message.warn("warn:"), msg, *args)


def print_hint(msg: str, *args):
    """
    Print a hint message in the style of the LLDB CLI.
    """
    print(message.hint("hint:"), msg, *args)


def print_info(msg: str, *args):
    """
    Print an information message in the style of the LLDB CLI.
    """
    print(message.info("info:"), msg, *args)


from pwndbg.dbg_mod.lldb.repl.proc import EventHandler
from pwndbg.dbg_mod.lldb.repl.proc import LaunchResultEarlyExit
from pwndbg.dbg_mod.lldb.repl.proc import LaunchResultError
from pwndbg.dbg_mod.lldb.repl.proc import ProcessDriver

show_tip = pwndbg.config.add_param(
    "show-tips", True, "whether to display the tip of the day on startup"
)

# We only allow certain commands to be executed in LLDB mode. This list contains
# tuples made up of the full name of the command and functions that check if a
# given command matches it.
LLDB_EXCLUSIVE = [
    ("script", lambda cmd: cmd.startswith("sc") and "script".startswith(cmd)),
    ("expression", lambda cmd: cmd.startswith("e") and "expression".startswith(cmd)),
]


def lex_args(args: str) -> list[str]:
    """
    Splits the arguments, respecting quotation marks.
    """
    args = args.strip()
    result = []
    while len(args) > 0:
        first = re.match("\\s*(\".*\"|'.*'|\\S+)", args)
        sl = first[1]

        # Handle single and double quotes, we could do some escaping for the
        # double quotes case, but we don't, yet.
        sl = sl.strip('"')
        sl = sl.strip("'")

        result.append(sl)
        args = args[first.end() :]

    return result


class EventRelay(EventHandler):
    """
    The event system that is sensible for the REPL process driver to use isn't
    an exact match with the one used by the rest of Pwndbg. They're close, but
    there's a bit of work we have to do to properly convey certain events.
    """

    def __init__(self, dbg: LLDB) -> None:
        self.dbg = dbg
        self.ignore_resumed = 0

    def _set_ignore_resumed(self, count: int) -> None:
        """
        Don't relay next given number of resumed events.
        """
        self.ignore_resumed += count

    @override
    def created(self) -> None:
        self.dbg._trigger_event(EventType.START)

    @override
    def suspended(self, event: lldb.SBEvent) -> None:
        # Under some circumstances - like when debugging core files - the
        # process driver may not actually know what the event is.
        if event is not None:
            # The event might have originated from a different source than the user
            # currently has selected. Move focus to the where the event happened.
            #
            # state-changed events have no thread associated with them, and so
            # SBThread::GetThreadFromEvent does not work. Interrogate each thread in
            # the process and look for the most interesting one.
            proc = lldb.SBProcess.GetProcessFromEvent(event)
            for thread in proc.threads:
                # Currently the one considered most interesting is simply the first
                # that has any reason at all to be stopped.
                if thread.stop_reason == lldb.eStopReasonNone:
                    continue

                if proc.GetSelectedThread().idx != thread.idx:
                    print(message.notice(f"[Switched to Thread {thread.id}]"))
                    assert proc.SetSelectedThread(thread)
                break

        self.dbg._trigger_event(EventType.STOP)

    @override
    def resumed(self) -> None:
        if self.ignore_resumed > 0:
            self.ignore_resumed -= 1
            return

        self.dbg._trigger_event(EventType.CONTINUE)

    @override
    def exited(self) -> None:
        self.dbg._trigger_event(EventType.EXIT)

    @override
    def modules_loaded(self) -> None:
        self.dbg._trigger_event(EventType.NEW_MODULE)


def show_greeting() -> None:
    """
    Show the Pwndbg greeting, the same way the GDB version of Pwndbg would. This
    one is considerably simpler than the GDB version, however, as we control the
    lifetime of the program, we know exactly when the greeting needs to be shown,
    so we don't bother with any of the lifetime checks.
    """
    hint_lines = (f"loaded {len(pwndbg.commands.commands)} pwndbg commands commands.",)

    for line in hint_lines:
        print(message.prompt("pwndbg: ") + message.system(line))

    if show_tip:
        colored_tip = color_tip(get_tip_of_the_day(pwndbg.dbg_mod.DebuggerType.LLDB.value))
        print(
            message.prompt("------- tip of the day (some of these don't work in LLDB yet!)")
            + message.system(" (disable with {})".format(message.notice("set show-tips off")))
            + message.prompt(" -------")
        )
        print(colored_tip)


class YieldExecDirect:
    """
    Execute the given command directly, on behalf of the user.
    """

    def __init__(self, command: str, capture: bool, prompt_silent: bool):
        self._command = command
        self._capture = capture
        self._prompt_silent = prompt_silent


class YieldInteractive:
    """
    Prompt the user for the next command.
    """

    pass


class PwndbgController:
    """
    Class providing interfaces for a client to control the behavior of Pwndbg
    asynchronously.
    """

    def interactive(self) -> Awaitable[None]:
        """
        Runs a single interactive round, in which the user is prompted for a
        command from standard input and `readline`, and whatever command they
        type in is executed.
        """
        return OneShotAwaitable(YieldInteractive())

    def execute(self, command: str) -> Awaitable[None]:
        """
        Runs the given command, and displays its output to the user.

        # Interactivity
        Some commands - such as `lldb` and `ipi` - start interactive prompts
        when they are run, and issuing them through this command will not change
        that behavior.
        """
        return OneShotAwaitable(YieldExecDirect(command, False, False))

    def execute_and_capture(self, command: str) -> Awaitable[bytes]:
        """
        Runs the given command, and captures its output as a byte string.

        # Interactivity
        Same caveats apply as in `execute`.

        # Reliabily of Capture
        Some Pwndbg commands currently do not have their outputs captured, even
        when run through this command. It is expected that this will be improved
        in the future, but, as as general rule, clients should not rely on the
        output of the command being available.
        """
        return OneShotAwaitable(YieldExecDirect(command, True, False))


@wrap_with_history
def run(
    controller: Callable[..., Coroutine[Any, Any, None]],
    *args,
    debug: bool = False,
) -> None:
    """
    Runs the Pwndbg CLI through the given asynchronous controller.
    """

    assert isinstance(pwndbg.dbg, LLDB)
    dbg: LLDB = pwndbg.dbg

    if HAS_FZF:
        session = get_prompt_session(dbg)
    else:
        enable_readline(dbg)

    # We're gonna be dealing with process events ourselves, so we'll want to run
    # LLDB in asynchronous mode.
    dbg.debugger.SetAsync(True)

    # This is the driver we're going to be using to handle the process.
    relay = EventRelay(dbg)
    with ProcessDriver(debug=debug, event_handler=relay) as driver:
        # Set ourselves up to respond to changes in Python execution context.
        curr_state = LLDBPythonState.PWNDBG
        in_lldb_command_handler = False

        def handle_lldb_python_state_change(new_state: LLDBPythonState) -> None:
            nonlocal curr_state
            nonlocal in_lldb_command_handler
            match (curr_state, new_state):
                case (LLDBPythonState.PWNDBG, LLDBPythonState.LLDB_STOP_HANDLER):
                    driver.pause_io_if_running()
                case (LLDBPythonState.PWNDBG, LLDBPythonState.LLDB_COMMAND_HANDLER):
                    driver.pause_io_if_running()
                case (_, LLDBPythonState.PWNDBG):
                    driver.resume_io_if_running()
            curr_state = new_state

        dbg.lldb_python_state_callback = handle_lldb_python_state_change

        # Set ourselves up to respond to SIGINT by interrupting the process if it is
        # running, and doing nothing otherwise.
        def handle_sigint(_sig, _frame):
            driver.interrupt(in_lldb_command_handler=in_lldb_command_handler)

        signal.signal(signal.SIGINT, handle_sigint)

        show_greeting()
        last_command = ""

        coroutine = controller(PwndbgController(), *args)
        last_result: Any = None
        last_exc: Exception | None = None

        while True:
            try:
                # Execute the prompt hook.
                dbg._fire_prompt_hook()

                try:
                    if last_exc is not None:
                        action = coroutine.throw(last_exc)
                    else:
                        action = coroutine.send(last_result)
                except StopIteration:
                    # Nothing else for us to do.
                    break
                except asyncio.CancelledError:
                    # We requested a cancellation that wasn't overwritten.
                    break
                finally:
                    last_exc = None
                    last_result = None

                if isinstance(action, YieldInteractive):
                    if debug:
                        print("[-] REPL: Prompt next command from user interactively")

                    try:
                        if HAS_FZF:
                            try:
                                line = session.prompt(message=PROMPT)
                            except KeyboardInterrupt:
                                continue
                        else:
                            line = input(PROMPT)
                        # If the input is empty (i.e., 'Enter'), use the previous command
                        if line:
                            last_command = line
                        else:
                            line = last_command
                    except EOFError:
                        # Exit the REPL if there's nothing else to run.
                        last_exc = asyncio.CancelledError()
                        continue

                    if not exec_repl_command(line, sys.stdout, dbg, driver, relay):
                        last_exc = asyncio.CancelledError()
                        continue

                elif isinstance(action, YieldExecDirect):
                    if debug:
                        print(
                            f"[-] REPL: Executing command '{action._command}' {'with' if action._capture else 'without'} output capture"
                        )

                    last_command = action._command

                    if not action._prompt_silent:
                        print(f"{PROMPT.value if HAS_FZF else PROMPT}{action._command}")

                    try:
                        if action._capture:
                            with TextIOWrapper(BytesIO(), write_through=True) as output:
                                should_continue = exec_repl_command(
                                    action._command, output, dbg, driver, relay
                                )
                                last_result = output.buffer.getvalue()
                                print(last_result.decode("utf-8", errors="surrogateescape"))
                        else:
                            should_continue = exec_repl_command(
                                action._command, sys.stdout, dbg, driver, relay
                            )
                    except BaseException as e:
                        last_exc = e
                        continue

                    if not should_continue:
                        last_exc = asyncio.CancelledError()
                        continue
            except UserCancelledError as e:
                last_exc = e


def exec_repl_command(
    line: str,
    output_to,
    dbg: LLDB,
    driver: ProcessDriver,
    relay: EventRelay,
) -> bool:
    """
    Parses and runs the given command, returning whether the event loop should continue.
    """
    stdout = None
    stderr = None
    lldb_out = None
    lldb_err = None
    try:
        stdout = sys.stdout
        stderr = sys.stderr
        lldb_out = dbg.debugger.GetOutputFile()
        lldb_err = dbg.debugger.GetErrorFile()

        sys.stdout = output_to
        dbg.debugger.SetOutputFile(
            lldb.SBFile.Create(output_to, borrow=True, force_io_methods=True)
        )
        dbg.debugger.SetErrorFile(lldb.SBFile.Create(output_to, borrow=True, force_io_methods=True))

        return _exec_repl_command(line, output_to.buffer, dbg, driver, relay)
    finally:
        if stdout is not None:
            sys.stdout = stdout
        if stderr is not None:
            sys.stderr = stderr
        if lldb_out is not None:
            dbg.debugger.SetOutputFile(lldb_out)
        if lldb_err is not None:
            dbg.debugger.SetErrorFile(lldb_err)


def _exec_repl_command(
    line: str,
    lldb_out_target: BinaryIO,
    dbg: LLDB,
    driver: ProcessDriver,
    relay: EventRelay,
) -> bool:
    """
    Implementation for exec_repl_command
    """

    bits = lex_args(line)

    if len(line) == 0:
        return True

    # Let the user get an LLDB prompt if they so desire.
    if bits[0] == "lldb":
        print_warn("You are now entering LLDB mode. To exit, type 'quit', 'exit' or Ctrl-D.")
        print_warn(
            "In this mode, certain commands may cause Pwndbg to break. Proceed with caution."
        )
        dbg.debugger.RunCommandInterpreter(
            True, False, lldb.SBCommandInterpreterRunOptions(), 0, False, False
        )
        return True

    # Not allowed to have this be a regular command, because of LLDB.
    if "version".startswith(line) and line.startswith("ve"):
        pwndbg.commands.version.version_impl()
        return True

    # There are interactive commands that `SBDebugger.HandleCommand` will
    # silently ignore. We have to implement them manually, here.
    if "quit".startswith(line) and line.startswith("q"):
        return False
    if "exit".startswith(line) and line.startswith("exi"):
        return False

    # `script` is a little weird. Unlike with the other commands we're
    # emulating, we actually need LLDB to spawn it for it to make sense
    # from the perspective of the user. This means we have to make
    # special arrangements for it.
    #
    # There is a way to get LLDB to properly handle interactive commands,
    # and that is to start an interactive session with
    # `SBDebugger.RunCommandInterpreter`, but that comes with its own
    # challenges:
    #     (1) Starting an interactive session on standard input is the
    #         best option from the perspective of the user, as they get
    #         full access to the Python interpreter's readline functions.
    #         However, we can't start a session running a command, which
    #         means we open up the possibility of the user breaking
    #         Pwndbg completely if they type in any process or target
    #         management commands.
    #     (2) Setting an input file up for the debugger to use, having
    #         that input file start the Python interpreter, and piping
    #         `sys.stdin` to it while the interpreter is running. This
    #         option is better in that it avoids the possibility of the
    #         user breaking Pwndbg by mistake, but it breaks both
    #         readline and input in general for the user.
    #
    # While neither option is ideal, both can be partially mitigated.
    # Option (1) by adding an extra command that drops down to LLDB and
    # prints a warning to make the user aware of the risk of breaking
    # Pwndbg, and option (2) by making a TextIOBase class that uses input()
    # at the REPL level before piping that to the Python interpreter running
    # under LLDB.
    #
    # Currently, we go with the mitigated version of option (1), but option
    # (2) might still be on the table for the near future.
    #
    # Likewise for the other commands we barr here.
    found_barred = False
    for name, test in LLDB_EXCLUSIVE:
        if not test(bits[0]):
            continue

        print_error(
            f"The '{name}' command is not supported. Use the 'lldb' command to enter LLDB mode and try again."
        )

        found_barred = True

    if found_barred:
        return True

    # Because we need to capture events related to target setup and process
    # startup, we handle them here, in a special way.
    if bits[0].startswith("pr") and "process".startswith(bits[0]):
        if len(bits) > 1 and bits[1].startswith("la") and "launch".startswith(bits[1]):
            # This is `process launch`.
            process_launch(driver, relay, bits[2:], dbg)
            return True
        if len(bits) > 1 and bits[1].startswith("a") and "attach".startswith(bits[1]):
            # This is `process attach`.
            process_attach(driver, relay, bits[2:], dbg)
            return True
        if len(bits) > 1 and bits[1].startswith("conn") and "connect".startswith(bits[1]):
            # This is `process connect`.
            process_connect(driver, relay, bits[2:], dbg)
            return True
        # We don't care about other process commands..

    if (bits[0].startswith("at") and "attach".startswith(bits[0])) or (
        bits[0].startswith("_regexp-a") and "_regexp-attach".startswith(bits[0])
    ):
        # `attach` is an alias for `_regexp-attach`
        # (it is NOT an alias for `process attach` even if it may seem so!)
        attach(driver, relay, bits[1:], dbg)
        return True

    if bits[0].startswith("ta") and "target".startswith(bits[0]):
        if len(bits) > 1 and bits[1].startswith("c") and "create".startswith(bits[1]):
            # This is `target create`
            target_create(driver, bits[2:], dbg)
            return True
        if len(bits) > 1 and bits[1].startswith("de") and "delete".startswith(bits[1]):
            # This is `target delete`
            #
            # Currently, this check is here but it does nothing. We might
            # need to check for this, but I can't figure out what kind of
            # processing we should do for its arguments, so we do nothing.
            pass

    if bits[0].startswith("r") and "run".startswith(bits[0]):
        # `run` is an alias for `process launch -X true --`
        process_launch(driver, relay, ["-X", "true", "--"] + bits[1:], dbg, restart=True)
        return True

    if bits[0].startswith("sta") and "starti".startswith(bits[0]):
        # `starti` is like `run` but stops at the first instruction (stop-at-entry)
        process_launch(driver, relay, ["-s", "-X", "true", "--"] + bits[1:], dbg, restart=True)
        return True

    if bits[0] == "c" or (bits[0].startswith("con") and "continue".startswith(bits[0])):
        # Handle `continue` manually. While `ProcessDriver.run_lldb_command`
        # is more than capable of handling this command itself, there's no
        # need for it to. We know what the user wants, so we can fast-track
        # their request.
        continue_process(driver, bits[1:], dbg)
        return True

    if bits[0].startswith("gd") and "gdb-remote".startswith(bits[0]):
        # `gdb-remote` is almost the same as `process launch -p gdb-remote`,
        # but it does some additional changes to the URL, by prepending
        # "connect://" to it. So, from our pespective, it is a separate
        # command, even though it will also end up calling process_launch().
        gdb_remote(driver, relay, bits[1:], dbg)
        return True

    if bits[0] == "set":
        # We handle `set` as a command override. We do this so that users
        # may change Pwndbg-specific settings in the same way that they
        # would in GDB Pwndbg.
        #
        # The alternatives to this are either (1) use a proper command,
        # but that requires the process to already be running, and needs us
        # to use a name other than "set", or (2) add our settings to the
        # standard debugger settings mechanism, like we do in GDB, but LLDB
        # doesn't support that.
        warn = False
        if len(bits) != 3:
            print("Usage: set <name> <value>")
            warn = True
        else:
            param = pget(bits[1])
            if param is None:
                print_error(f"unknown setting '{bits[1]}'")
                warn = True
            else:
                try:
                    pset(param, bits[2])
                except InvalidParse as e:
                    print_error(f"invalid value '{bits[2]}' for setting '{bits[1]}': {e}")
                    warn = True

        if warn:
            print_hint(
                "Use the 'config', 'theme' and 'help set' commands to see what settings are available."
            )
            print_hint(
                "Use the 'help set <name>' command to see information about a specific setting."
            )
            print_hint(
                "If you meant to change LLDB settings, use the fully spelled-out 'settings' command, instead."
            )

        return True

    if (
        (bits[0] == "h" or (bits[0].startswith("hel") and "help".startswith(bits[0])))
        and len(bits) >= 2
        and bits[1] == "set"
    ):
        # This is 'help set'
        #
        # We override this command to provide help for the 'set' command
        # override.
        warn = False
        if len(bits) > 3:
            print("Usage: help set [name]")
            warn = True
        elif len(bits) == 2:
            # In LLDB style, list all valid settings.
            print("Set a Pwndbg configuration parameter.")
            print()
            print("Syntax: set <name> <value>")
            for scope in pwndbg.lib.config.Scope:
                print()
                print(f"Configuration parameters - {scope._name_}:")
                pwndbg.commands.config.display_config("", scope, show_hints=False)
        else:
            # Show information about a single parameter.
            param = pget(bits[2])
            if param is None:
                print_error(f"unknown setting '{bits[2]}'")
                warn = True
            else:
                print(f"Set {param.set_show_doc}.")
                print()
                print(f"Syntax: set {param.name} <value>")
                print()
                if param.help_docstring:
                    print(param.help_docstring)

        if warn:
            print_hint("Use the 'help set' command to see what settings are available for Pwndbg.")
            print_hint(
                "If you meant to see help for LLDB settings, see the fully spelled-out 'settings' command, instead."
            )

        return True

    if bits[0] == "ipi":
        # Spawn IPython shell, easy for debugging
        run_ipython_shell()
        return True

    if (
        bits[0] == pwndbg.commands.start.entry.command_name
        or bits[0] in pwndbg.commands.start.entry.aliases
    ):
        # 'entry' is actually a Pwndbg command. For convenience, we launch the
        # process on its behalf, before letting it run.
        #
        # In the LLDB back-end, there is no proper mechanism to make a process
        # start from inside of a command, as there is in GDB. Ideally, we'd
        # rework `ProcessDriver` so that it lets us do that with an execution
        # controller, but that is quite a bit of work to fix a single command,
        # when using an override is enough to achieve the same goal.
        #
        # In any case, we should consider doing it if this proves to be too janky.
        if not driver.has_process():
            process_launch(driver, relay, ["-s"], dbg)

        # This intentionally falls through. We want LLDB to do the rest of the
        # work of processing 'entry'.

    # The command hasn't matched any of our filtered commands, just let LLDB
    # handle it normally. Either in the context of the process, if we have
    # one, or just in a general context.
    if driver.has_process():
        driver.run_lldb_command(line, lldb_out_target)
        dbg.relay_exceptions()
    else:
        ret = lldb.SBCommandReturnObject()
        dbg.debugger.GetCommandInterpreter().HandleCommand(line, ret)
        if ret.IsValid():
            # LLDB can give us strings that may fail to encode.
            out = ret.GetOutput().strip()
            if len(out) > 0:
                lldb_out_target.write(out.encode(sys.stdout.encoding, errors="backslashreplace"))
                lldb_out_target.write(b"\n")
            out = ret.GetError().strip()
            if len(out) > 0:
                lldb_out_target.write(out.encode(sys.stdout.encoding, errors="backslashreplace"))
                lldb_out_target.write(b"\n")
        dbg.relay_exceptions()

    # At this point, the last command might've queued up some execution
    # control procedures for us to chew on. Run them now.
    coroutine_fail_warn = False
    for process, coroutine in dbg.controllers:
        assert driver.has_process()
        assert driver.process.GetUniqueID() == process.process.GetUniqueID()

        if driver.is_core_file():
            print_warn("skipping coroutine: target is core file")
            continue

        try:
            driver.run_coroutine(coroutine)
        except Exception:
            # We treat exceptions coming from the execution controllers the
            # same way we treat exceptions coming from commands.
            pwndbg.exception.handle()
            coroutine_fail_warn = True

    dbg.controllers.clear()

    if coroutine_fail_warn:
        print_warn(
            "Exceptions occurred execution controller processing. Debugging will likely be unreliable going forward."
        )

    return True


def _bool_of_string(val: str) -> bool:
    """
    Convert a string to a boolean value.

    For use with ArgumentParser.
    """
    if val.lower() in ("true", "1", "yes"):
        return True
    if val.lower() in ("false", "0", "no"):
        return False
    raise ValueError(f"{val} is not a recognized boolean value")


def parse(
    args: list[str],
    parser: argparse.ArgumentParser,
    unsupported: list[str],
    raw_marker: str | None = None,
) -> Any | None:
    """
    Parses a list of string arguments into an object containing the parsed
    data.

    If `raw_marker` is not `None`, the argument list will be split in
    two, with all arguments before the split being fed to the argument parser,
    and all arguments after the split being returned as-is. In this case the
    return value is a tuple.
    """
    raw = None
    if raw_marker is not None:
        # Always return something, even if we match nothing.
        raw = []

        try:
            index = args.index(raw_marker)
            raw = args[index + 1 :]
            args = args[:index]
        except ValueError:
            # Ugly, but avoids going over the list twice.
            pass

    try:
        args = parser.parse_args(args)
    except SystemExit:
        # Ugly, but need to keep ArgumentParser from terminating the process.
        return None

    # Reject any arguments we don't know how to handle yet.
    #
    # We'd like this list to grow over time, but we don't strictly need to
    # support all of these right away.
    varsargs = vars(args)
    for unsup in unsupported:
        if varsargs[unsup.replace("-", "_")]:
            print_error(f"Pwndbg does not support --{unsup} yet")
            return None

    if raw is not None:
        # If called with `raw_marked`, return a tuple.
        return args, raw

    return args


class AutoTarget:
    """
    During the execution of some commands, the LLDB CLI automatically creates an
    empty target and selects it before the command is executed.
    """

    def __init__(self, dbg: LLDB):
        self.error = lldb.SBError()
        self._dbg = dbg
        self._created_target = False

        count = dbg.debugger.GetNumTargets()
        if count == 0:
            # Create the target.
            self.target = dbg.debugger.CreateTarget(None, None, None, True, self.error)

            if not self.error.success:
                return

            # On success, select it and remember that it has been created.
            dbg.debugger.SetSelectedTarget(self.target)
            self._created_target = True
        elif count == 1:
            # Just use the current target.
            self.target = dbg.debugger.GetTargetAtIndex(0)
            assert self.target, (
                f"SBDebugger::GetNumTargets() is 1, but SBDebugger::GetTargetAtIndex(0) is {self.target}"
            )
        else:
            raise AssertionError(
                f"Pwndbg does not support multiple targets, so SBDebugger::GetNumTargets() must always be 0 or 1, but is {count}"
            )

    def __bool__(self):
        return self.error.success

    def close(self):
        if self._created_target:
            assert self._dbg.debugger.DeleteTarget(self.target), (
                "Could not delete the target we've just created. What?"
            )


def run_ipython_shell():
    @contextmanager
    def switch_to_ipython_env():
        saved_excepthook = sys.excepthook
        try:
            saved_ps = sys.ps1, sys.ps2
        except AttributeError:
            saved_ps = None
        yield
        # Restore Python's default `ps1`, `ps2`, and `excepthook`
        # to ensure proper behavior of the LLDB `script` command.
        if saved_ps is not None:
            sys.ps1, sys.ps2 = saved_ps
        else:
            del sys.ps1
            del sys.ps2
        sys.excepthook = saved_excepthook

    def start_ipi():
        import IPython
        import jedi  # type: ignore[import-untyped]

        jedi.Interpreter._allow_descriptor_getattr_default = False
        IPython.embed(
            colors="neutral", banner1="", confirm_exit=False, simple_prompt=False, user_ns=globals()
        )

    with switch_to_ipython_env():
        start_ipi()


target_create_ap = argparse.ArgumentParser(add_help=False, prog="target create")
target_create_ap.add_argument("-S", "--sysroot")
target_create_ap.add_argument("-a", "--arch")
target_create_ap.add_argument("-b", "--build")
target_create_ap.add_argument("-c", "--core", action="store_true")
target_create_ap.add_argument("-d", "--no-dependents")
target_create_ap.add_argument("-p", "--platform")
target_create_ap.add_argument("-r", "--remote-file")
target_create_ap.add_argument("-s", "--symfile")
target_create_ap.add_argument("-v", "--version")
target_create_ap.add_argument("filename")
target_create_unsupported = [
    "build",
    "no-dependents",
    "remote-file",
    "symfile",
    "version",
]


def _get_target_triple(debugger: lldb.SBDebugger, filepath: str) -> str | None:
    # The triple is the "architecture-vendor-OS[-ABI]" of the target binary.
    # Examples:
    # - "arm--linux-eabi"
    # - "aarch64--linux"
    # - "x86_64-apple-macosx11.7.0"
    # - "arm64-apple-macosx11.7.0"
    # - "aarch64-pc-windows-msvc"
    target: lldb.SBTarget = debugger.CreateTarget(filepath)
    if not target.IsValid():
        return None
    triple = target.triple
    debugger.DeleteTarget(target)
    return triple


def target_create(driver: ProcessDriver, args: list[str], dbg: LLDB) -> None:
    """
    Creates a new target, registers it with the Pwndbg LLDB implementation, and
    sets up listeners for it.
    """
    args = parse(args, target_create_ap, target_create_unsupported)
    if not args:
        return

    if args.core:
        target_create_core(driver, args, dbg)
    else:
        target_create_regular(args, dbg)


def target_create_core(driver: ProcessDriver, args: Any, dbg: LLDB) -> None:
    """
    Starts core file debugging.

    Despite the similar names and commands, this function is fundamentally
    different from `target_create_regular`. This is so we can match the way the
    regular LLDB CLI handles `target create --core`.

    In the regular LLDB CLI, when one types in `target create --core <name>`,
    a new special process is automatically launched, whose SBProcess object
    behaves differently from those of regular processes.

    In effect, in the LLDB CLI, creating a core file target both automatically
    starts the process and puts the CLI into a special core file mode, in which
    most lifetime-related user actions are forbidden.
    """
    if dbg.debugger.GetNumTargets() > 0:
        print_error(
            "Pwndbg does not support multiple targets. Please remove the current target with 'target delete' and try again."
        )
        return

    # Create a no-target SBTarget that will hold our core file.
    target = dbg.debugger.CreateTarget(None)
    if target is None or not target.IsValid():
        print_error("could not create target for core file debugging")
        return
    dbg.debugger.SetSelectedTarget(target)

    result = driver.launch_core_file(target, args.filename)
    match result:
        case LaunchResultError(what):
            print_error(f"could not launch core file: {what.description}")
            return

    split_triple = target.triple.split("-")
    if len(split_triple) > 0:
        arch = split_triple[0]
    else:
        arch = "unknown"

    print(f"Core file '{args.filename}' ({arch}) was loaded.")


def target_create_regular(args: Any, dbg: LLDB) -> None:
    """
    Creates a new regular target, and registers it with LLDB.
    """
    if dbg.debugger.GetNumTargets() > 0:
        print_error(
            "Pwndbg does not support multiple targets. Please remove the current target with 'target delete' and try again."
        )
        return

    if args.platform and args.platform not in {"qemu-user"}:
        print_error("Pwndbg does currently support platforms: qemu-user")
        return

    if args.arch:
        dbg.debugger.SetDefaultArchitecture(args.arch)

    if args.sysroot:
        dbg.debugger.SetCurrentPlatformSDKRoot(args.sysroot)

    # Create the target with the debugger.
    error = lldb.SBError()
    if args.platform:
        dbg.debugger.SetCurrentPlatform(args.platform)

        # Having the platform specified requires that we specify the triple.
        triple = _get_target_triple(dbg.debugger, args.filename)
        if not triple:
            print_error(f"could not detect triple for '{args.filename}'")
            return

        if args.platform == "qemu-user":
            arch = triple.split("-")[0]
            # Without setting it qemu-user don't work ;(
            dbg._execute_lldb_command(f"settings set platform.plugin.qemu-user.architecture {arch}")

        target: lldb.SBTarget = dbg.debugger.CreateTarget(
            args.filename, triple, args.platform, True, error
        )
    else:
        # Let LLDB figure out both the triple and the platform automatically.
        target = dbg.debugger.CreateTarget(args.filename, None, None, True, error)
    if not error.success or not target.IsValid():
        print_error(f"could not create target for '{args.filename}': {error.description}")
        return

    dbg.debugger.SetSelectedTarget(target)
    print(f"Current executable set to '{args.filename}' ({target.triple.split('-')[0]})")


process_launch_ap = argparse.ArgumentParser(add_help=False, prog="process launch")
process_launch_ap.add_argument("-A", "--disable-aslr", type=_bool_of_string, default=False)
process_launch_ap.add_argument("-C", "--script-class")
process_launch_ap.add_argument("-E", "--environment", action="append")
process_launch_ap.add_argument("-P", "--plugin")
process_launch_ap.add_argument("-X", "--shell-expand-args", type=_bool_of_string)
process_launch_ap.add_argument("-a", "--arch")
process_launch_ap.add_argument("-c", "--shell")
process_launch_ap.add_argument("-e", "--stderr")
process_launch_ap.add_argument("-i", "--stdin")
process_launch_ap.add_argument("-k", "--structured-data-key")
process_launch_ap.add_argument("-n", "--no-stdio")
process_launch_ap.add_argument("-o", "--stdout")
process_launch_ap.add_argument("-s", "--stop-at-entry", action="store_true")
process_launch_ap.add_argument("-t", "--tty")
process_launch_ap.add_argument("-v", "--structured-data-value")
process_launch_ap.add_argument("-w", "--working-dir")
process_launch_ap.add_argument("run-args", nargs="*")
process_launch_unsupported = [
    "script-class",
    "plugin",
    "arch",
    "shell",
    "stderr",
    "stdin",
    "structured-data-key",
    "no-stdio",
    "stdout",
    "tty",
    "structured-data-value",
    "working-dir",
]


def process_launch(
    driver: ProcessDriver, relay: EventRelay, args: list[str], dbg: LLDB, restart: bool = False
) -> None:
    """
    Launches a process with the given arguments.
    If restart is True, kills any existing process first.
    """
    result = parse(args, process_launch_ap, process_launch_unsupported, raw_marker="--")
    if result is None:
        return
    args, raw = result

    launch_args = getattr(args, "run-args", []) + raw
    if args.shell_expand_args:
        # Perform shell expansion.
        launch_args = [os.path.expanduser(os.path.expandvars(arg)) for arg in launch_args]

    targets = dbg.debugger.GetNumTargets()
    assert targets < 2
    if targets == 0:
        print_error("no target, create one using the 'target create' command")
        return

    if driver.has_process():
        if restart:
            print_info("killing running process and starting a fresh one")
            what = driver.kill()
            if not what.success:
                print_error(f"could not kill existing process: {result.description}")
                return
        else:
            print_error("a process is already being debugged")
            return

    target: lldb.SBTarget = dbg.debugger.GetTargetAtIndex(0)

    # Make sure LLDB knows the correct remote or local status of this launch.
    dbg._current_process_is_gdb_remote = driver.has_connection()

    if target.GetPlatform().GetName() == "qemu-user":
        # Force qemu-user as remote, pwndbg depends on that, eg: for download procfs files
        dbg._current_process_is_gdb_remote = True

    io_driver = get_io_driver()
    result = driver.launch(
        target,
        io_driver,
        [f"{name}={value}" for name, value in os.environ.items()]
        + (args.environment if args.environment else []),
        launch_args,
        os.getcwd(),
        args.disable_aslr,
    )

    match result:
        case LaunchResultError(what, disconnected):
            print_error(f"could not launch process: {what.description}")
            if disconnected:
                print_warn("disconnected")
            return
        case LaunchResultEarlyExit():
            print_warn("process exited early")
            if sys.platform == "darwin":
                print_hint(
                    "It may be that you are trying to launch a protected binary on a SIP enabled system."
                )
            return

    # Continue execution if the user hasn't requested for a stop at the entry
    # point of the process. And handle necessary events.
    if not args.stop_at_entry:
        # The relay has already sended a START event at this point. Continuing
        # normally will send a CONTINUE event, which is incorrect, as START
        # already implies the program is about to start running. So we avoid
        # sending the next CONTINUE event.
        relay._set_ignore_resumed(1)

        driver.cont()
    else:
        # Tell the debugger that the process was suspended.
        #
        # The event system in the process driver can't natively represent the
        # START event type exactly. It knows only when a process has been
        # created and when a process changed it state to running from suspended,
        # and vice versa.
        #
        # This means that we have to relay an extra event here to convey that
        # the process stopped at entry, even though what's going on, in reality,
        # is that we're simply not resuming the process.
        dbg._trigger_event(EventType.STOP)


process_attach_ap = argparse.ArgumentParser(add_help=False, prog="process attach")
process_attach_ap.add_argument("-C", "--python-class")
process_attach_ap.add_argument("-P", "--plugin")
process_attach_ap.add_argument("-c", "--continue", action="store_true")
process_attach_ap.add_argument("-i", "--include-existing", action="store_true")
process_attach_ap.add_argument("-k", "--structured-data-key")
process_attach_ap.add_argument("-n", "--name")
process_attach_ap.add_argument("-p", "--pid", type=int)
process_attach_ap.add_argument("-v", "--structured-data-value")
process_attach_ap.add_argument("-w", "--waitfor", action="store_true")
process_attach_unsupported = [
    "python-class",
    "plugin",
    "structured-data-key",
    "structured-data-value",
]


def _attach_with_info(
    driver: ProcessDriver, relay: EventRelay, dbg: LLDB, info: lldb.SBAttachInfo, cont=False
):
    """
    Attaches to a process based on SBAttachInfo information
    """
    assert dbg.debugger.GetNumTargets() < 2

    # TODO/FIXME: This should ask:
    # 'There is a running process, detach from it and attach?: [Y/n]'
    if driver.has_process():
        print_error("a process is already being debugged")
        return

    auto = AutoTarget(dbg)
    if not auto:
        print_error(f"could not create empty target for attaching: {auto.error.description}")
        auto.close()
        return

    # Make sure LLDB knows the correct remote or local status of this attach.
    dbg._current_process_is_gdb_remote = driver.has_connection()

    result = driver.attach(
        auto.target,
        info,
    )

    match result:
        case LaunchResultError(what, disconnected):
            print_error(f"could not attach to process: {what.description}")
            if disconnected:
                print_warn("disconnected")
            auto.close()
            return
        case LaunchResultEarlyExit():
            print_warn("process exited early")
            if sys.platform == "darwin":
                print_hint(
                    "It may be that you are trying to attach to a protected binary on a SIP enabled system."
                )
            auto.close()
            return

    # Continue execution if the user has requested it.
    if cont:
        # Same logic applies here as in `process_launch`.
        relay._set_ignore_resumed(1)
        driver.cont()
    else:
        # Same logic applies here as in `process_launch`.
        dbg._trigger_event(EventType.STOP)


def process_attach(driver: ProcessDriver, relay: EventRelay, args: list[str], dbg: LLDB) -> None:
    """
    Attaches to a process with the given arguments.
    """
    args = parse(args, process_attach_ap, process_attach_unsupported)
    if not args:
        return

    # The first two arguments - executable name and wait_for_launch - don't
    # matter, we set them later. The third one is required, as it tells LLDB the
    # attach should be asynchronous.
    info = lldb.SBAttachInfo(None, False, True)

    if args.name is not None:
        info.SetExecutable(args.name)
    if args.pid is not None:
        info.SetProcessID(args.pid)
    info.SetWaitForLaunch(args.waitfor)

    do_continue = getattr(args, "continue", False)
    if do_continue:
        info.SetResumeCount(1)
    info.SetIgnoreExisting(not args.include_existing)

    _attach_with_info(driver, relay, dbg, info, cont=do_continue)


def attach(driver: ProcessDriver, relay: EventRelay, args: list[str], dbg: LLDB) -> None:
    """
    Attaches to a process with the given name or pid based on regex match.
    Used for `_regexp-attach <pid|name>` (alias for `attach <pid|name>`)
    Note: for some reason, `attach` does not really take a regex for process name.
    """
    if len(args) != 1:
        print_error("expected 1 argument: <pid> or <name>")
        return
    arg = args[0]

    # exec name - None, we set it later (or pid)
    # wait_for_launch - False, since we don't wait
    # third arg - tell LLDB to attach asynchronously
    info = lldb.SBAttachInfo(None, False, True)

    # Argument is pid
    if arg.isdigit():
        info.SetProcessID(int(arg))
    else:
        info.SetExecutable(arg)

    _attach_with_info(driver, relay, dbg, info)


process_connect_ap = argparse.ArgumentParser(add_help=False, prog="process connect")
process_connect_ap.add_argument("-p", "--plugin")
process_connect_ap.add_argument("remoteurl")


def process_connect(driver: ProcessDriver, relay: EventRelay, args: list[str], dbg: LLDB) -> None:
    """
    Connects to the given remote process.
    """
    args = parse(args, process_connect_ap, [])
    if not args:
        return

    if "plugin" not in args or args.plugin != "gdb-remote":
        print_error(
            "Pwndbg only supports the gdb-remote plugin for 'process connect'. Please specify it with the '-p gdb-remote' argument."
        )
        return

    if driver.has_connection():
        print_error("debugger is already connected")
        return

    # Make sure the LLDB driver knows that this is a remote process.
    dbg._current_process_is_gdb_remote = True

    auto = AutoTarget(dbg)
    if not auto:
        print_error(f"could not create empty target for connection: {auto.error.description}")
        auto.close()
        return

    io_driver = get_io_driver()
    result = driver.connect(auto.target, io_driver, args.remoteurl, "gdb-remote")

    match result:
        case LaunchResultError(what, disconnected):
            print_error(f"could not connect to remote: {what.description}")
            if disconnected:
                print_warn("disconnected")
            auto.close()
            return
        case LaunchResultEarlyExit():
            print_warn("remote exited early")
            auto.close()
            return

    # Tell the debugger that the process was suspended, if there is a process.
    if driver.has_process():
        dbg._trigger_event(EventType.STOP)


gdb_remote_ap = argparse.ArgumentParser(add_help=False, prog="gdb-remote")
gdb_remote_ap.add_argument("remoteurl")


def gdb_remote(driver: ProcessDriver, relay: EventRelay, args: list[str], dbg: LLDB) -> None:
    """
    Like `process_connect`, but more lenient with the remote URL format.
    """

    args = parse(args, gdb_remote_ap, [])
    if not args:
        return

    parts = args.remoteurl.split(":")
    if len(parts) == 1:
        url = None
        port = parts[0]
    elif len(parts) == 2:
        url = parts[0]
        port = parts[1]
    else:
        print_error(f"unknown URL format '{args.remoteurl}'")
        return

    try:
        port = int(port, 10)
    except ValueError:
        print_error(f"could not interpret '{port}' as port number")
        return

    if url is None:
        print_warn("hostname not given, using 'localhost'")
        url = "localhost"

    process_connect(driver, relay, ["-p", "gdb-remote", f"connect://{url}:{port}"], dbg)


continue_ap = argparse.ArgumentParser(add_help=False, prog="continue")
continue_ap.add_argument("-i", "--ignore-count")
continue_unsupported = ["ignore-count"]


def continue_process(driver: ProcessDriver, args: list[str], dbg: LLDB) -> None:
    """
    Continues the execution of a process.
    """
    args = parse(args, continue_ap, continue_unsupported)
    if not args:
        return

    if not driver.has_process():
        print_error("no process")
        return

    if driver.is_core_file():
        print_error("cannot continue: target is core file")
        return

    driver.cont()



"""
Common types.
"""

from __future__ import annotations

import sys

import pwndbg
import pwndbg.dbg_mod

module = sys.modules[__name__]

char: pwndbg.dbg_mod.Type
ulong: pwndbg.dbg_mod.Type
long: pwndbg.dbg_mod.Type
uchar: pwndbg.dbg_mod.Type
ushort: pwndbg.dbg_mod.Type
uint: pwndbg.dbg_mod.Type
sint: pwndbg.dbg_mod.Type
void: pwndbg.dbg_mod.Type

uint8: pwndbg.dbg_mod.Type
uint16: pwndbg.dbg_mod.Type
uint32: pwndbg.dbg_mod.Type
uint64: pwndbg.dbg_mod.Type
unsigned: dict[int, pwndbg.dbg_mod.Type]

int8: pwndbg.dbg_mod.Type
int16: pwndbg.dbg_mod.Type
int32: pwndbg.dbg_mod.Type
int64: pwndbg.dbg_mod.Type
signed: dict[int, pwndbg.dbg_mod.Type]

pvoid: pwndbg.dbg_mod.Type
ppvoid: pwndbg.dbg_mod.Type
pchar: pwndbg.dbg_mod.Type

ptrsize: int = 4

ptrdiff: pwndbg.dbg_mod.Type
size_t: pwndbg.dbg_mod.Type
ssize_t: pwndbg.dbg_mod.Type


def lookup_types(*types: str) -> pwndbg.dbg_mod.Type:
    try:
        process = pwndbg.dbg.selected_inferior()
    except pwndbg.dbg_mod.NoInferior:
        raise AssertionError("Tried to initialize typeinfo with no inferior")
    for type_str in types:
        t = process.types_with_name(type_str)
        if len(t) > 0:
            return t[0]

    return None


def update() -> None:
    module.char = lookup_types("signed char", "char")
    module.ulong = lookup_types("unsigned long", "uint", "u32", "uint32")
    module.long = lookup_types("long", "int", "i32", "int32")
    module.uchar = lookup_types("unsigned char", "ubyte", "u8", "uint8")
    module.ushort = lookup_types("unsigned short", "ushort", "u16", "uint16", "uint16_t")
    module.uint = lookup_types("unsigned int", "uint", "u32", "uint32")
    # Putting 'signed int' at the front is slow for kernel. Not sure how it's possible
    # that type is missing. See PR #3115.
    module.sint = lookup_types("int", "signed int", "signed")
    module.void = lookup_types("void", "()")

    module.uint8 = module.uchar
    module.uint16 = module.ushort
    module.uint32 = module.uint
    module.uint64 = lookup_types("unsigned long long", "ulong", "u64", "uint64")
    module.unsigned = {
        1: module.uint8,
        2: module.uint16,
        4: module.uint32,
        8: module.uint64,
    }

    module.int8 = lookup_types("char", "i8", "int8")
    module.int16 = lookup_types("short", "short int", "i16", "int16")
    module.int32 = lookup_types("int", "i32", "int32")
    module.int64 = lookup_types("long long", "long long int", "long", "i64", "int64")
    module.signed = {1: module.int8, 2: module.int16, 4: module.int32, 8: module.int64}

    module.pvoid = void.pointer()
    module.ppvoid = pvoid.pointer()
    module.pchar = char.pointer()

    module.ptrsize = pvoid.sizeof

    if pvoid.sizeof == 4:
        module.ptrdiff = module.uint32
        module.size_t = module.uint32
        module.ssize_t = module.int32
    elif pvoid.sizeof == 8:
        module.ptrdiff = module.uint64
        module.size_t = module.uint64
        module.ssize_t = module.int64
    else:
        raise Exception("Pointer size not supported")


def load(name: str) -> pwndbg.dbg_mod.Type | None:
    """Load a symbol; note that new symbols can be added with `add-symbol-file` functionality"""
    names = pwndbg.dbg.selected_inferior().types_with_name(name)
    if len(names) > 0:
        return names[0]
    return None


def get_type(size: int) -> pwndbg.dbg_mod.Type:
    return {
        1: pwndbg.aglib.typeinfo.uint8,
        2: pwndbg.aglib.typeinfo.uint16,
        4: pwndbg.aglib.typeinfo.uint32,
        8: pwndbg.aglib.typeinfo.uint64,
    }[size]



"""
Routines to enumerate mapped memory, and attempt to associate
address ranges with various ELF files and permissions.

The reason that we need robustness is that not every operating
system has /proc/$$/maps, which backs 'info proc mapping'.
"""

from __future__ import annotations

from collections.abc import Iterator

import gdb

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.elf
import pwndbg.aglib.file
import pwndbg.aglib.proc
import pwndbg.aglib.qemu
import pwndbg.auxv
import pwndbg.gdblib.info
import pwndbg.lib.cache
import pwndbg.lib.memory


@pwndbg.lib.cache.cache_until("objfile", "start")
def is_corefile() -> bool:
    """
    For example output use:
        gdb ./tests/binaries/crash_simple.out -ex run -ex 'generate-core-file ./core' -ex 'quit'

    And then use:
        gdb ./tests/binaries/crash_simple.out -core ./core -ex 'info target'
    And:
        gdb -core ./core

    As the two differ in output slighty.
    """
    inf = gdb.selected_inferior()
    conn = inf.connection
    if conn is None:
        return False
    if not isinstance(inf.connection, gdb.TargetConnection):
        return False
    return conn.type == "core"


@pwndbg.lib.cache.cache_until("start", "stop")
def get_known_maps() -> tuple[pwndbg.lib.memory.Page, ...] | None:
    """
    Similar to `vmmap.get()`, except only returns maps in cases where
    the mappings are known, like if it's a coredump, or if process
    mappings are available.
    """
    # Note: debugging a coredump does still show proc.alive() == True
    if not pwndbg.aglib.proc.alive():
        return ()

    if is_corefile():
        return tuple(coredump_maps())

    return proc_tid_maps()


def iter_coredump_sections() -> Iterator[pwndbg.lib.memory.Page]:
    lines = gdb.execute("maintenance info sections", to_string=True)
    started_sections = False
    ptrsize = pwndbg.aglib.arch.ptrsize

    for line in lines.splitlines():
        if not started_sections:
            if "Core file:" in line:
                started_sections = True
            continue

        if not line.startswith(" "):
            # End of section
            break

        # We look for lines like:
        # ['[9]', '0x00000000->0x00000150', 'at', '0x00098c40:', '.auxv', 'HAS_CONTENTS']
        # ['[15]', '0x555555555000->0x555555556000', 'at', '0x00001430:', 'load2', 'ALLOC', 'LOAD', 'READONLY', 'CODE', 'HAS_CONTENTS']
        try:
            _idx, start_end, _at_str, _at, section_name, *flags_list = line.split()
            start, end = (int(v, 16) for v in start_end.split("->"))

            # Skip pages with start=0x0, this is unlikely this is valid vmmap
            if start == 0:
                continue

            # Tried taking this from the 'at 0x...' value
            # but it turns out to be invalid, so keep it 0 until we find better way
            offset = 0
        except (IndexError, ValueError):
            continue

        # Note: can we deduce anything from 'ALLOC', 'HAS_CONTENTS' or 'LOAD' flags?
        flags = 0
        if "READONLY" in flags_list:
            flags |= 4
        if "DATA" in flags_list:
            flags |= 2
        if "CODE" in flags_list:
            flags |= 1

        page = pwndbg.lib.memory.Page(
            start, end - start, flags, offset, ptrsize, f"[{section_name}]"
        )
        yield page


def enhance_coredump_sections_pages_info(pages: list[pwndbg.lib.memory.Page]) -> None:
    for section in iter_coredump_sections():
        # Now, if the section is already in pages, just add its perms
        known_page = False

        for page in pages:
            if section.start in page:
                page.flags |= section.flags
                known_page = True
                break

        if known_page:
            continue

        pages.append(section)


def enhance_known_pages_info(pages: list[pwndbg.lib.memory.Page]) -> None:
    if not pages:
        return

    # ffffffffff600000-ffffffffff601000
    # If the last page starts on e.g. 0xffffffffff600000 it must be vsyscall
    vsyscall_page = pages[-1]
    if vsyscall_page.start > 0xFFFFFFFFFF000000 and vsyscall_page.flags & 1:
        vsyscall_page.objfile = "[vsyscall]"
        vsyscall_page.offset = 0

    # Detect stack based on addresses in AUXV from stack memory
    stack_addr = None
    vdso_addr = None

    # TODO/FIXME: Can we uxe `pwndbg.auxv.get()` for this somehow?
    auxv = pwndbg.gdblib.info.auxv().splitlines()
    for line in auxv:
        if "AT_EXECFN" in line:
            try:
                stack_addr = int(line.split()[-2], 16)
            except Exception:
                pass
        if "AT_SYSINFO_EHDR" in line:
            try:
                vdso_addr = int(line.split()[-1], 16)
            except Exception:
                pass

    for page in pages:
        if stack_addr and stack_addr in page:
            page.objfile = "[stack]"
            page.flags |= 6
            page.offset = 0
        if vdso_addr:
            if vdso_addr == page.start:
                page.objfile = "[vdso]"
            elif vdso_addr == page.end:
                page.objfile = "[maybe: vvar/vvar_vclock]"


@pwndbg.lib.cache.cache_until("objfile", "start")
def coredump_maps() -> tuple[pwndbg.lib.memory.Page, ...]:
    """
    Parses `info proc mappings` and `maintenance info sections`
    and tries to make sense out of the result :)
    """
    pages = list(info_proc_maps(parse_flags=False))

    enhance_coredump_sections_pages_info(pages)
    enhance_known_pages_info(pages)

    pages.sort(key=lambda page: page.start)
    return tuple(pages)


def parse_info_proc_mappings_line(
    line: str, perms_available: bool, parse_flags: bool
) -> pwndbg.lib.memory.Page | None:
    """
    Parse a line from `info proc mappings` and return a pwndbg.lib.memory.Page
    object if the line is valid.

    Example lines:
        0x4c3000           0x4c5000     0x2000    0xc2000  rw-p   /root/hello_world/main
        0x4c5000           0x4cb000     0x6000        0x0  rw-p

    The objfile column might be empty, and the permissions column is only present in GDB versions >= 12.1
    https://github.com/bminor/binutils-gdb/commit/29ef4c0699e1b46d41ade00ae07a54f979ea21cc

    Args:
        line: A line from `info proc mappings`.

    Returns:
        A pwndbg.lib.memory.Page object or None.
    """
    try:
        # Example line with all fields present: ['0x555555555000', '0x555555556000', '0x1000', '0x1000', 'rw-p', '/home/user/a.out']
        split_line = line.split()

        start_str = split_line[0]
        _end = split_line[1]
        size_str = split_line[2]
        offset_str = split_line[3]

        if perms_available:
            perm = split_line[4]
            # The objfile column may be empty.
            objfile = split_line[5] if len(split_line) > 5 else ""
        else:
            perm = "rwxp"
            objfile = split_line[4] if len(split_line) > 4 else ""

        start, size, offset = int(start_str, 16), int(size_str, 16), int(offset_str, 16)
    except (IndexError, ValueError):
        return None

    flags = 0
    if parse_flags:
        if "r" in perm:
            flags |= 4
        if "w" in perm:
            flags |= 2
        if "x" in perm:
            flags |= 1

    ptrsize = pwndbg.aglib.arch.ptrsize
    return pwndbg.lib.memory.Page(start, size, flags, offset, ptrsize, objfile)


@pwndbg.lib.cache.cache_until("start", "stop")
def info_proc_maps(parse_flags: bool = True) -> tuple[pwndbg.lib.memory.Page, ...]:
    """
    Parse the result of info proc mappings.

    Example output:

            Start Addr           End Addr       Size     Offset  Perms  objfile
              0x400000           0x401000     0x1000        0x0  r--p   /root/hello_world/main
              0x401000           0x497000    0x96000     0x1000  r-xp   /root/hello_world/main
              0x497000           0x4be000    0x27000    0x97000  r--p   /root/hello_world/main
              0x4be000           0x4c3000     0x5000    0xbd000  r--p   /root/hello_world/main
              0x4c3000           0x4c5000     0x2000    0xc2000  rw-p   /root/hello_world/main
              0x4c5000           0x4cb000     0x6000        0x0  rw-p
              0x4cb000           0x4ed000    0x22000        0x0  rw-p   [heap]
        0x7ffff7ff9000     0x7ffff7ffd000     0x4000        0x0  r--p   [vvar]
        0x7ffff7ffd000     0x7ffff7fff000     0x2000        0x0  r-xp   [vdso]
        0x7ffffffde000     0x7ffffffff000    0x21000        0x0  rw-p   [stack]
    0xffffffffff600000 0xffffffffff601000     0x1000        0x0  --xp   [vsyscall]

    Note: this may return no pages due to a bug/behavior of GDB.
    See https://sourceware.org/bugzilla/show_bug.cgi?id=31207
    for more information.

    Returns:
        A tuple of pwndbg.lib.memory.Page objects or an empty tuple if
        info proc mapping is not supported on the target.
    """

    try:
        info_proc_mappings = pwndbg.gdblib.info.proc_mappings().splitlines()
    except gdb.error:
        # On qemu user emulation, we may get: gdb.error: Not supported on this target.
        info_proc_mappings = []

    # See if "Perms" is in the header line
    perms_available = len(info_proc_mappings) >= 4 and "Perms" in info_proc_mappings[3]

    pages: list[pwndbg.lib.memory.Page] = []
    for line in info_proc_mappings:
        page = parse_info_proc_mappings_line(line, perms_available, parse_flags)
        if page is not None:
            pages.append(page)

    return tuple(pages)


def parse_tid_maps_line(line: str) -> pwndbg.lib.memory.Page:
    # Example /proc/$tid/maps
    # 7f95266fa000-7f95268b5000 r-xp 00000000 08:01 418404                     /lib/x86_64-linux-gnu/libc-2.19.so
    # 7f95268b5000-7f9526ab5000 ---p 001bb000 08:01 418404                     /lib/x86_64-linux-gnu/libc-2.19.so
    # 7f9526ab5000-7f9526ab9000 r--p 001bb000 08:01 418404                     /lib/x86_64-linux-gnu/libc-2.19.so
    # 7f9526ab9000-7f9526abb000 rw-p 001bf000 08:01 418404                     /lib/x86_64-linux-gnu/libc-2.19.so
    # 7f9526abb000-7f9526ac0000 rw-p 00000000 00:00 0
    # 7f9526ac0000-7f9526ae3000 r-xp 00000000 08:01 418153                     /lib/x86_64-linux-gnu/ld-2.19.so
    # 7f9526cbe000-7f9526cc1000 rw-p 00000000 00:00 0
    # 7f9526ce0000-7f9526ce2000 rw-p 00000000 00:00 0
    # 7f9526ce2000-7f9526ce3000 r--p 00022000 08:01 418153                     /lib/x86_64-linux-gnu/ld-2.19.so
    # 7f9526ce3000-7f9526ce4000 rw-p 00023000 08:01 418153                     /lib/x86_64-linux-gnu/ld-2.19.so
    # 7f9526ce4000-7f9526ce5000 rw-p 00000000 00:00 0
    # 7f9526ce5000-7f9526d01000 r-xp 00000000 08:01 786466                     /bin/dash
    # 7f9526f00000-7f9526f02000 r--p 0001b000 08:01 786466                     /bin/dash
    # 7f9526f02000-7f9526f03000 rw-p 0001d000 08:01 786466                     /bin/dash
    # 7f9526f03000-7f9526f05000 rw-p 00000000 00:00 0
    # 7f95279fe000-7f9527a1f000 rw-p 00000000 00:00 0                          [heap]
    # 7fff3c177000-7fff3c199000 rw-p 00000000 00:00 0                          [stack]
    # 7fff3c1e8000-7fff3c1ea000 r-xp 00000000 00:00 0                          [vdso]
    # ffffffffff600000-ffffffffff601000 r-xp 00000000 00:00 0                  [vsyscall]
    ptrsize: int = pwndbg.aglib.arch.ptrsize
    maps, perm, offset, dev, inode_objfile = line.split(maxsplit=4)

    start, stop = maps.split("-")

    try:
        inode, objfile = inode_objfile.split(maxsplit=1)
    except Exception:
        # Name unnamed anonymous pages so they can be used e.g. with search commands
        objfile = "[anon_" + start[:-3] + "]"

    start = int(start, 16)
    stop = int(stop, 16)
    offset = int(offset, 16)
    size = stop - start

    flags = 0
    if "r" in perm:
        flags |= 4
    if "w" in perm:
        flags |= 2
    if "x" in perm:
        flags |= 1

    return pwndbg.lib.memory.Page(start, size, flags, offset, ptrsize, objfile)


def parse_tid_maps(data: str) -> list[pwndbg.lib.memory.Page]:
    pages: list[pwndbg.lib.memory.Page] = []
    for line in data.splitlines():
        page = parse_tid_maps_line(line)
        pages.append(page)

    return pages


def parse_tid_smaps_dict(data: list[str]) -> dict[str, list[str]]:
    smaps_dict: dict[str, list[str]] = {}
    for line in data:
        try:
            key, *value = line.strip().split()
            key = key.rstrip(":")
            smaps_dict[key.replace(":", "")] = value
        except ValueError:
            continue

    return smaps_dict


def group_tid_smaps_segments(data: str) -> list[list[str]]:
    # Example segment of /proc/$tid/smaps
    # 7f0bd1c25000-7f0bd1c27000 rw-p 001e8000 00:22 1655096                    /usr/lib64/libc.so.6
    # Size:                  8 kB
    # KernelPageSize:        4 kB
    # MMUPageSize:           4 kB
    # Rss:                   8 kB
    # Pss:                   8 kB
    # Pss_Dirty:             8 kB
    # Shared_Clean:          0 kB
    # Shared_Dirty:          0 kB
    # Private_Clean:         0 kB
    # Private_Dirty:         8 kB
    # Referenced:            8 kB
    # Anonymous:             8 kB
    # KSM:                   0 kB
    # LazyFree:              0 kB
    # AnonHugePages:         0 kB
    # ShmemPmdMapped:        0 kB
    # FilePmdMapped:         0 kB
    # Shared_Hugetlb:        0 kB
    # Private_Hugetlb:       0 kB
    # Swap:                  0 kB
    # SwapPss:               0 kB
    # Locked:                0 kB
    # THPeligible:           0
    # ProtectionKey:         0
    # VmFlags: rd wr mr mw me ac sd
    segments: list[list[str]] = []
    lines = data.splitlines()

    # Group lines into segments by detecting address range lines
    current_segment: list[str] = []

    for line in lines:
        # Check if this line starts with an address range (e.g., "7f0bd1c25000-7f0bd1c27000")
        # An address range line starts with hex digits followed by a dash
        is_address_line = False
        if line and not line[0].isspace():
            # Try to detect if this is an address range line
            try:
                first_token = line.split()[0]
                if "-" in first_token:
                    # Try to parse both parts as hex addresses
                    start, end = first_token.split("-", 1)
                    int(start, 16)
                    int(end, 16)
                    is_address_line = True
            except (ValueError, IndexError):
                pass

        if is_address_line:
            # Process the previous segment if it exists
            if current_segment:
                segments.append(current_segment)

            # Start a new segment
            current_segment = [line]
        else:
            # Add to current segment
            current_segment.append(line)

    # Process the last segment
    if current_segment:
        segments.append(current_segment)

    return segments


def parse_tid_smaps(data: str) -> list[pwndbg.lib.memory.Page]:
    pages: list[pwndbg.lib.memory.Page] = []
    segments = group_tid_smaps_segments(data)
    for segment in segments:
        page = parse_tid_maps_line(segment[0])
        smaps_dict = parse_tid_smaps_dict(segment[1:])
        if "ProtectionKey" in smaps_dict:
            page.protection_key = int(smaps_dict.get("ProtectionKey")[0])
        page.vm_flags = smaps_dict.get("VmFlags")
        pages.append(page)

    return pages


@pwndbg.lib.cache.cache_until("start", "stop")
def proc_tid_maps() -> tuple[pwndbg.lib.memory.Page, ...] | None:
    """
    Parse the contents of /proc/$TID/smaps on the server.
    Falls back to /proc/$TID/maps if smaps is not available.
    (TID == Thread Identifier. We do not use PID since it may not be correct)

    Returns:
        A tuple of pwndbg.lib.memory.Page objects or None if
        /proc/$tid/maps doesn't exist or when we debug a qemu-user target
    """

    # If we debug remotely a qemu-system target,
    # there is no point of hitting things further
    if pwndbg.aglib.qemu.is_qemu_kernel():
        return None

    tid = pwndbg.aglib.proc.tid()
    locations = [
        # Linux distro
        (f"/proc/{tid}/smaps", parse_tid_smaps),
        (f"/proc/{tid}/maps", parse_tid_maps),
        # Freebsd in some cases
        (f"/usr/compat/linux/proc/{tid}/maps", parse_tid_maps),
    ]
    maps_parse_function = None

    for location, parse_fn in locations:
        try:
            data = pwndbg.aglib.file.get(location).decode()
            maps_parse_function = parse_fn
            break
        except OSError:
            continue
    else:
        return None

    # Process hasn't been fully created yet; it is in Z (zombie) state
    if data == "":
        return ()

    pages = maps_parse_function(data)

    return tuple(pages)



"""
Reading register value from the inferior, and provides a
standardized interface to registers like "sp" and "pc".
"""

from __future__ import annotations

import ctypes
import re
from collections.abc import Generator
from collections.abc import Iterator
from typing import Any

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.proc
import pwndbg.aglib.remote
import pwndbg.dbg_mod
import pwndbg.lib.cache
from pwndbg.dbg_mod import EventType
from pwndbg.lib.regs import BitFlags
from pwndbg.lib.regs import KernelRegisterSet
from pwndbg.lib.regs import RegisterSet
from pwndbg.lib.regs import reg_sets

# We need to manually make some ptrace calls to get fs/gs bases on Intel
PTRACE_ARCH_PRCTL = 30
ARCH_GET_FS = 0x1003
ARCH_GET_GS = 0x1004


class RegisterManager:
    previous: dict[str, int | None] = {}
    last: dict[str, int | None] = {}

    @pwndbg.lib.cache.cache_until("stop")
    def regs_in_frame(self, frame: pwndbg.dbg_mod.Frame) -> pwndbg.dbg_mod.Registers:
        return frame.regs()

    @pwndbg.aglib.proc.OnlyWhenRunning
    def get_register(self, name: str, frame: pwndbg.dbg_mod.Frame) -> pwndbg.dbg_mod.Value | None:
        regs = self.regs_in_frame(frame)
        value = regs.by_name(name)
        return value if value is not None else regs.by_name(name.upper())

    @pwndbg.aglib.proc.OnlyWhenQemuKernel
    @pwndbg.aglib.proc.OnlyWhenRunning
    def get_qemu_register(self, name: str) -> int | None:
        out = pwndbg.dbg.selected_inferior().send_monitor("info registers")
        match = re.search(rf"{name.split('_')[0]}=\s+([\da-fA-F]+)\s+([\da-fA-F]+)", out)

        if match:
            base = int(match.group(1), 16)
            limit = int(match.group(2), 16)

            if name.endswith("LIMIT"):
                return limit
            return base

        return None

    def read_reg_uncached_in_frame(self, reg: str, frame: pwndbg.dbg_mod.Frame) -> int | None:
        reg = reg.lstrip("$")
        try:
            value = self.get_register(reg, frame)
            if value is None and reg.lower() == "xpsr":
                value = self.get_register("xPSR", frame)
            if value is None:
                return None
            value = int(value)
            if reg == "eip" and pwndbg.aglib.arch.name == "i8086":
                cs = self.get_register("cs", frame)
                if cs is None:
                    return None
                value += int(cs) * 0x10

            # The value that the native debugger returns can be negative.
            # We convert this to the unsigned bit representation by masking it
            reg_definition = pwndbg.aglib.regs.current.reg_definitions.get(reg.lower())
            if reg_definition and reg_definition.mask is not None:
                mask = reg_definition.mask
            else:
                mask = pwndbg.aglib.arch.ptrmask
            return int(value) & mask
        except (ValueError, pwndbg.dbg_mod.Error):
            return None

    def read_reg_uncached(self, reg: str) -> int | None:
        frame = pwndbg.dbg.selected_frame()
        if frame is None:
            return None
        return self.read_reg_uncached_in_frame(reg, frame)

    @pwndbg.lib.cache.cache_until("stop")
    def read_reg_in_frame(self, reg: str, frame: pwndbg.dbg_mod.Frame) -> int | None:
        """
        Same as read_reg() except for the provided frame, rather than the currently
        selected frame.
        """
        return self.read_reg_uncached_in_frame(reg, frame)

    def read_reg(self, reg: str) -> int | None:
        """
        Query the underlying debugger for the value of a register.

        Note that in some rare cases, debuggers won't directly expose the values of some special model specific registers.
        Although we can sometimes determine these by other indirect means, this function does not run any extra logic to handle these special cases.

        Specifically, if you need to ensure you are reading the correct value of "gs", "fs", "idt", or "idt_limit", use
        the specific helpers functions on the regs module as necessary to determine the values.

        Use read_reg_in_frame() if you have a `frame` object, its faster.
        """
        # Adding a cache_until decorator to read_reg() is semantically incorrect since it will return
        # the same register value even if the frame changes.
        frame = pwndbg.dbg.selected_frame()
        if frame is None:
            return None
        return self.read_reg_in_frame(reg, frame)

    def write_reg(self, reg: str, value: int) -> None:
        if not pwndbg.dbg.selected_frame().reg_write(reg, value):
            raise RuntimeError(f"Attempted to write to a non-existent register '{reg}'")

    @property
    def pc(self) -> int | None:
        """Get the value of the program counter register"""
        return self.read_reg(self.current.pc)

    @pc.setter
    def pc(self, val: int) -> None:
        self.write_reg(self.current.pc, val)

    @property
    def sp(self) -> int | None:
        """Get the value of the stack pointer register"""
        return self.read_reg(self.current.stack)

    @sp.setter
    def sp(self, val: int) -> None:
        """Get the value of the stack pointer register"""
        self.write_reg(self.current.stack, val)

    def __contains__(self, reg: str) -> bool:
        return reg_sets[pwndbg.aglib.arch.name].__contains__(reg)

    def __iter__(self) -> Iterator[str]:
        return reg_sets[pwndbg.aglib.arch.name].__iter__()

    @property
    def current(self) -> RegisterSet:
        return reg_sets[pwndbg.aglib.arch.name]

    # TODO: All these should be able to do self.current
    @property
    def gpr(self) -> tuple[str, ...]:
        return reg_sets[pwndbg.aglib.arch.name].gpr

    @property
    def common(self) -> list[str]:
        return reg_sets[pwndbg.aglib.arch.name].common

    @property
    def frame(self) -> str | None:
        return reg_sets[pwndbg.aglib.arch.name].frame

    @property
    def retaddr(self) -> tuple[str, ...]:
        return reg_sets[pwndbg.aglib.arch.name].retaddr

    @property
    def kernel(self) -> KernelRegisterSet | None:
        return reg_sets[pwndbg.aglib.arch.name].kernel

    @property
    def flags(self) -> dict[str, BitFlags]:
        return reg_sets[pwndbg.aglib.arch.name].flags

    @property
    def extra_flags(self) -> dict[str, BitFlags]:
        return reg_sets[pwndbg.aglib.arch.name].extra_flags

    @property
    def stack(self) -> str:
        return reg_sets[pwndbg.aglib.arch.name].stack

    @property
    def retval(self) -> str | None:
        return reg_sets[pwndbg.aglib.arch.name].retval

    @property
    def all(self) -> set[str]:
        return reg_sets[pwndbg.aglib.arch.name].all

    def fix(self, expression: str) -> str:
        """
        This is used in CLI parsing.
        It takes in a string with a register name, "rax", and prefixes it with
        a $ ("$rax") so that the underlying debugger can evaluate it to resolve the value
        """
        expression = pwndbg.aglib.regs.current.resolve_aliases(expression)
        for regname in self.all:
            expression = re.sub(rf"\$?\b{regname}\b", r"$" + regname, expression)
        return expression

    def items(self) -> Generator[tuple[str, Any], None, None]:
        for regname in self.all:
            yield regname, self.read_reg(regname)

    reg_sets = reg_sets

    @property
    def changed(self) -> list[str]:
        delta: list[str] = []
        for reg, value in self.previous.items():
            if self.read_reg(reg) != value:
                delta.append(reg)
        return delta

    @property
    @pwndbg.aglib.proc.OnlyWhenQemuKernel
    @pwndbg.aglib.proc.OnlyWithArch(["i386", "x86-64"])
    @pwndbg.lib.cache.cache_until("stop")
    def idt(self) -> int | None:
        return self.get_qemu_register("IDT")

    @property
    @pwndbg.aglib.proc.OnlyWhenQemuKernel
    @pwndbg.aglib.proc.OnlyWithArch(["i386", "x86-64"])
    @pwndbg.lib.cache.cache_until("stop")
    def idt_limit(self) -> int | None:
        return self.get_qemu_register("IDT_LIMIT")

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def fsbase(self) -> int:
        return self._fs_gs_helper("fs_base", ARCH_GET_FS)

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def gsbase(self) -> int:
        return self._fs_gs_helper("gs_base", ARCH_GET_GS)

    def _fs_gs_helper(self, regname: str, which: int) -> int:
        """Supports fetching based on segmented addressing, a la fs:[0x30].
        Requires ptrace'ing the child directory if i386."""

        if pwndbg.aglib.arch.name == "x86-64":
            frame = pwndbg.dbg.selected_frame()
            if frame is None:
                return 0
            reg_value = self.get_register(regname, frame)
            return int(reg_value) if reg_value is not None else 0

        # We can't really do anything if the process is remote.
        if pwndbg.aglib.remote.is_remote():
            return 0

        # Use the lightweight process ID
        lwpid = pwndbg.dbg.selected_thread().ptid()

        # Get the register
        ppvoid = ctypes.POINTER(ctypes.c_void_p)
        value = ppvoid(ctypes.c_void_p())
        value.contents.value = 0

        libc = ctypes.CDLL("libc.so.6")
        result = libc.ptrace(PTRACE_ARCH_PRCTL, lwpid, value, which)

        if result == 0:
            return (value.contents.value or 0) & pwndbg.aglib.arch.ptrmask

        return 0


regs: RegisterManager = RegisterManager()


@pwndbg.dbg.event_handler(EventType.CONTINUE)
@pwndbg.dbg.event_handler(EventType.STOP)
def update_last() -> None:
    regs.previous = regs.last
    regs.last = {k: regs.read_reg(k) for k in regs.common}
    # TODO: Uncomment this once the LLDB command port PR for `context` is merged
    # if pwndbg.config.show_retaddr_reg:
    #    M.last.update({k: M[k] for k in M.retaddr})



from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable
from collections.abc import Sequence
from enum import Enum
from functools import total_ordering
from typing import Any
from typing import TypeVar

T = TypeVar("T")

# Boolean value. True or False, same as in Python.
PARAM_BOOLEAN = 0
# Boolean value, or 'auto'.
PARAM_AUTO_BOOLEAN = 1
# Signed integer value. Disallows zero.
PARAM_INTEGER = 2
# Signed integer value.
PARAM_ZINTEGER = 3
# Unsigned integer value. Disallows zero.
PARAM_UINTEGER = 4
# Unsigned integer value.
PARAM_ZUINTEGER = 5
# Unlimited ZUINTEGER.
PARAM_ZUINTEGER_UNLIMITED = 6
# String value. Accepts escape sequences.
PARAM_STRING = 7
# String value, accepts only one of a number of possible values, specified at
# parameter creation.
PARAM_ENUM = 8
# String value corresponding to the name of a file, if present.
PARAM_OPTIONAL_FILENAME = 9

PARAM_CLASSES = {
    # The Python boolean values, True and False are the only valid values.
    bool: PARAM_BOOLEAN,
    # This is like PARAM_INTEGER, except 0 is interpreted as itself.
    int: PARAM_ZINTEGER,
    # When the user modifies the string, any escape sequences,
    # such as ‘\t’, ‘\f’, and octal escapes, are translated into
    # corresponding characters and encoded into the current host charset.
    str: PARAM_STRING,
}

# Strings that will be used in supplementing the parameter help_docstring
HELP_DEFAULT_PREFIX = "Default:"
HELP_VALID_VALUES_PREFIX = "Valid values:"


class Scope(Enum):
    # If you want to add another scope here, don't forget to add
    # a command which prints it!
    config = 1
    theme = 2
    heap = 3


# @total_ordering allows us to implement `__eq__` and `__lt__` and have all the
# other comparison operators handled for us
@total_ordering
class Parameter:
    def __init__(
        self,
        name: str,
        default: Any,
        set_show_doc: str,
        *,
        help_docstring: str = "",
        param_class: int | None = None,
        enum_sequence: Sequence[str] | None = None,
        scope: Scope = Scope.config,
    ) -> None:
        self.name = name
        self.default = default
        self._value = default
        self.param_class = param_class or PARAM_CLASSES[type(default)]
        # Note: If `set_show_doc` is "the value of foo" then:
        # The `set_doc` will be "Set the value of foo."
        # The `show_doc` will be "Show the value of foo."
        # `get_set_string()` will return "Set the value of foo to VALUE."
        # `get_show_string()` will return "Show the value of foo."
        self.set_show_doc = set_show_doc
        self.help_docstring = help_docstring
        # Show the default value in the parameter help.
        # We add a trailing double space because docs are in markdown.
        self.help_docstring += "\n\n" + HELP_DEFAULT_PREFIX + " " + self.pretty_default() + "  "
        # Show valid values if they aren't obvious
        if param_class == PARAM_ENUM:
            self.help_docstring += (
                "\n"
                + HELP_VALID_VALUES_PREFIX
                + " "
                + ", ".join([f"'{name}'" for name in enum_sequence])
            )
        if param_class == PARAM_AUTO_BOOLEAN:
            self.help_docstring += "\n" + HELP_VALID_VALUES_PREFIX + " on, off, auto."

        self.enum_sequence = enum_sequence
        self.scope = scope
        self.update_listeners: list[Callable[[Any], None]] = []

    def add_update_listener(self, listener: Callable[[Any], None]) -> None:
        self.update_listeners.append(listener)

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        self._value = value
        for listener in self.update_listeners:
            listener(value)

    @property
    def is_changed(self) -> bool:
        return self.value != self.default

    def revert_default(self) -> None:
        self.value = self.default

    def attr_name(self) -> str:
        """Returns the attribute name associated with this config option,
        i.e. `my-config` has the attribute name `my_config`"""
        return self.name.replace("-", "_")

    def __getattr__(self, name: str):
        return getattr(self.value, name)

    def pretty_val(self, val: Any) -> str:
        """Convert a value this object could contain to its pretty string representation."""
        if self.param_class == PARAM_BOOLEAN:
            return "on" if val else "off"
        if self.param_class == PARAM_AUTO_BOOLEAN:
            if val is None:
                return "auto"
            return "on" if val else "off"
        if (
            self.param_class == PARAM_STRING
            or self.param_class == PARAM_ENUM
            or self.param_class == PARAM_OPTIONAL_FILENAME
        ):
            return "'" + val + "'"
        return str(val)

    def pretty(self) -> str:
        return self.pretty_val(self.value)

    def pretty_default(self) -> str:
        return self.pretty_val(self.default)

    # Casting
    def __int__(self) -> int:
        return int(self.value)

    def __str__(self) -> str:
        return str(self.value)

    def __bool__(self) -> bool:
        return bool(self.value)

    # Compare operators
    # Ref: http://portingguide.readthedocs.io/en/latest/comparisons.html

    # If comparing with another `Parameter`, the `Parameter` objects are equal
    # if they refer to the same GDB parameter. For any other type of object, the
    # `Parameter` is equal to the object if `self.value` is equal to the object
    def __eq__(self, other: object) -> bool:
        if isinstance(other, Parameter):
            return self.name == other.name
        return self.value == other

    def __lt__(self, other: object) -> bool:
        if isinstance(other, Parameter):
            return self.name < other.name
        return self.value < other

    # Operators
    def __add__(self, other: int) -> int:
        return self.value + other

    def __radd__(self, other: int) -> int:
        return other + self.value

    def __sub__(self, other: int) -> int:
        return self.value - other

    def __rsub__(self, other: int) -> int:
        return other - self.value

    def __mul__(self, other: int) -> int:
        return self.value * other

    def __rmul__(self, other: int) -> str:
        return other * self.value

    def __div__(self, other: float) -> float:
        return self.value / other

    def __floordiv__(self, other: int) -> int:
        return self.value // other

    def __pow__(self, other: int) -> int:
        return self.value**other

    def __mod__(self, other: int) -> int:
        return self.value % other

    def __len__(self) -> int:
        return len(self.value)


class Config:
    def __init__(self) -> None:
        self.params: dict[str, Parameter] = {}
        self.triggers: defaultdict[str, list[Callable[..., Any]]] = defaultdict(list)

    def add_param(
        self,
        name: str,
        default: Any,
        set_show_doc: str,
        *,
        help_docstring: str = "",
        param_class: int | None = None,
        enum_sequence: Sequence[str] | None = None,
        scope: Scope = Scope.config,
    ) -> Parameter:
        # Dictionary keys are going to have underscores, so we can't allow them here
        assert "_" not in name
        assert len(name) <= 32 and "Parameter name too long."
        assert name and "Parameter name cannot be empty."

        set_show_doc = set_show_doc.strip()
        help_docstring = help_docstring.strip()
        assert set_show_doc and "Parameter's set_show_doc cannot be empty."
        assert (
            len(set_show_doc) <= 70
            and "Parameter's set_show_doc too long, use the help_docstring parameter."
        )
        assert set_show_doc[-1] != "." and "Don't end set_show_doc with punctuation."
        assert (
            HELP_DEFAULT_PREFIX not in help_docstring
            and f"Having the string '{HELP_DEFAULT_PREFIX}' in the help_docstring "
            "messes with documentation generation. Please remove it, it is automatically generated."
        )
        assert (
            HELP_VALID_VALUES_PREFIX not in help_docstring
            and f"Having the string '{HELP_VALID_VALUES_PREFIX}' in the help_docstring "
            "messes with documentation generation. Please remove it, you can use param_class=PARAM_ENUM."
        )

        if param_class == PARAM_ENUM or enum_sequence:
            assert param_class == PARAM_ENUM and enum_sequence

        assert scope in Scope

        p = Parameter(
            name,
            default,
            set_show_doc,
            help_docstring=help_docstring,
            param_class=param_class,
            enum_sequence=enum_sequence,
            scope=scope,
        )
        return self.add_param_obj(p)

    def add_param_obj(self, p: Parameter) -> Parameter:
        attr_name = p.attr_name()

        # Make sure this isn't a duplicate parameter
        assert attr_name not in self.params

        self.params[attr_name] = p
        return p

    def trigger(self, *params: Parameter) -> Callable[[Callable[..., T]], Callable[..., T]]:
        names = [p.name for p in params]

        def wrapper(func: Callable[..., T]) -> Callable[..., T]:
            for name in names:
                self.triggers[name].append(func)
            return func

        return wrapper

    def get_params(self, scope: Scope) -> list[Parameter]:
        return sorted(filter(lambda p: p.scope == scope, self.params.values()))

    def __getattr__(self, name: str) -> Parameter:
        if name in self.params:
            return self.params[name]
        raise AttributeError(f"'Config' object has no attribute '{name}'")

    def __setattr__(self, attr, val):
        if attr in ("params", "triggers"):
            return super().__setattr__(attr, val)
        raise AttributeError("Use config.<param>.value to set value of a parameter")



from __future__ import annotations

from dataclasses import dataclass

import pwndbg.aglib.memory


@dataclass
class SavedRegisterFrame:
    """
    A list of registers that have been saved to process memory for later restoration.

    For example, on syscall entry, the process registers are saved to the kernel stack.
    """

    # List of (offset, register name), sorted from smallest to largest offset
    frame_layout: list[tuple[int, str]]
    offsets: dict[str, int]

    def __init__(self, register_offsets: dict[str, int]):
        self.offsets = register_offsets

        self.frame_layout = sorted((y, x) for (x, y) in register_offsets.items())

    def read_saved_register(self, reg: str, sp: int = None) -> int | None:
        if sp is None:
            sp = pwndbg.aglib.regs.sp

        try:
            mem = pwndbg.aglib.memory.read(sp + self.offsets[reg], pwndbg.aglib.arch.ptrsize)
        except pwndbg.dbg_mod.Error:
            return None

        return pwndbg.aglib.arch.unpack(mem)


# Basic exception stack frame defined here - https://developer.arm.com/documentation/107706/0100/Exceptions-and-interrupts-overview/Stack-frames
ARM_CORTEX_M_EXCEPTION_STACK_FRAME_OFFSETS = {
    "r0": 0x0,
    "r1": 0x4,
    "r2": 0x8,
    "r3": 0xC,
    "r12": 0x10,
    "lr": 0x14,
    "pc": 0x18,
    "xpsr": 0x1C,
}


ARM_CORTEX_M_EXCEPTION_STACK = SavedRegisterFrame(ARM_CORTEX_M_EXCEPTION_STACK_FRAME_OFFSETS)



from __future__ import annotations

import re
from enum import Enum
from random import choice

from pwndbg.color import message

# GDB specific tips
GDB_TIPS: list[str] = [
    "GDB's `apropos <topic>` command displays all registered commands that are related to the given <topic>",
    "GDB's `follow-fork-mode` parameter can be used to set whether to trace parent or child after fork() calls. Pwndbg sets it to child by default",
    'Use GDB\'s `dprintf` command to print all calls to given function. E.g. `dprintf malloc, "malloc(%p)\\n", (void*)$rdi` will print all malloc calls',
    "Use GDB's `pi` command to run an interactive Python console where you can use Pwndbg APIs like `pwndbg.aglib.memory.read(addr, len)`, `pwndbg.aglib.memory.write(addr, data)`, `pwndbg.aglib.vmmap.get()` and so on!",
    "GDB's `set directories <path>` parameter can be used to debug e.g. glibc sources like the malloc/free functions!",
    "If you have debugging symbols the `info args` command shows current frame's function arguments (use `up` and `down` to switch between frames)",
    'Calling functions like `call (void)puts("hello world")` will run all other target threads for the time the function runs. Use `set scheduler-locking on` to lock the execution to current thread when calling functions',
    "Use the `pipe <cmd> | <prog>` command to pass output of a GDB/Pwndbg command to a shell program, e.g. `pipe elfsections | grep bss`. This can also be shortened to: `| <cmd> | <prog>`",
    "Prefixing a command with `!` in GDB will execute it as a shell command, e.g.: `!ls` or `!cat flag.txt`",
]

# Pwndbg specific tips
PWNDBG_TIPS: list[str] = [
    "If you want Pwndbg to clear screen on each command (but still save previous output in history) use `set context-clear-screen on`",
    "The `set show-flags on` setting will display CPU flags register in the regs context panel",
    "GDB and Pwndbg parameters can be shown or set with `show <param>` and `set <param> <value>` GDB commands",
    "Use Pwndbg's `config` and `theme` commands to tune its configuration and theme colors!",
    "Pwndbg mirrors some of WinDbg commands like `eq`, `ew`, `ed`, `eb`, `es`, `dq`, `dw`, `dd`, `db`, `ds` for writing and reading memory",
    "Pwndbg resolves kernel memory maps by parsing page tables (default) or via `monitor info mem` QEMU gdbstub command (use `set kernel-vmmap-via-page-tables off` for that)",
    "Use the `vmmap` command for a better & colored memory maps display (than the GDB's `info proc mappings`)",
    "Use the `telescope` command to dereference a given address/pointer multiple times (if the dereferenced value is a valid ptr; see `config telescope` to configure its behavior)",
    "Use the `context` (or `ctx`) command to display the context once again. You can reconfigure the context layout with `set context-section <sections>` or forward the output to a file/tty via `set context-output <file>`. See also `config context` to configure it further!",
    "Disable Pwndbg context information display with `set context-sections ''`",
    "Pwndbg context displays where the program branches to thanks to emulating few instructions into the future. You can disable this with `set emulate off` which may also speed up debugging",
    "Use the `canary` command to see all stack canary/cookie values on the stack (based on the *usual* stack canary value initialized by glibc)",
    "Use the `procinfo` command for better process introspection (than the GDB's `info proc` command)",
    "Want to display each context panel in a separate tmux window? See https://pwndbg.re/stable/tutorials/splitting-the-context/",
    'Use `$base("heap")` to get the start address of a [heap] memory page',
    "Use the `errno` (or `errno <number>`) command to see the name of the last or provided (libc) error",
    "Pwndbg sets the SIGLARM, SIGBUS, SIGPIPE and SIGSEGV signals so they are not passed to the app; see `info signals` for full GDB signals configuration",
    "Use `vmmap -A|-B <number> <filter>` to display <number> of maps after/before filtered ones",
    "Use the `killall` command to kill all specified threads (via their ids)",
    "Use the `spray` command to spray memory with cyclic pattern or specified value",
    "Use `patch <address> '<assembly>'` to patch an address with given assembly code",
    "Want to NOP some instructions? Use `patch <address> 'nop; nop; nop'`",
    "`heap-config` shows heap related configuration",
    "`break-if-taken` and `break-if-not-taken` commands sets breakpoints after a given jump instruction was taken or not",
    "`stepuntilasm <assembly-instruction [operands]>` steps program forward until matching instruction occures",
    "Use `plist` command to dump elements of linked list",
    "If your program has multiple threads they will be displayed in the context display or using the `context threads` command",
    "Use `track-got enable|info|query` to track GOT accesses - useful for hijacking control flow via writable GOT/PLT",
    "Need to `mmap` or `mprotect` memory in the debugee? Use commands with the same name to inject and run such syscalls",
    "Use `hi` to see if a an address belongs to a glibc heap chunk",
    "Use `contextprev` and `contextnext` to display a previous context output again without scrolling",
    "Try splitting the context output into multiple TUI windows using `layout pwndbg` (`tui disable` or `ctrl-x + a` to go back to CLI mode)",
]

# LLDB specific tips
LLDB_TIPS: list[str] = [
    "Use LLDB's `help <command>` to get detailed help on any command",
    "LLDB's `expr` command lets you evaluate expressions in the current frame context",
    "Use `frame variable` (or `fr v`) to show all variables in the current frame",
    "The `watchpoint set` command allows you to stop execution when a variable changes",
    "Use `process launch --stop-at-entry` to stop at the program entry point",
    "LLDB's `memory read` (or `m read`) command displays memory contents at a specified address",
    "Use `thread backtrace all` to see backtraces of all threads",
    "The `breakpoint set --func-regex <regex>` command sets breakpoints on functions matching a regular expression",
    "Use `target modules list` to see all loaded modules in your process",
    "LLDB's `image lookup` command helps find symbols, addresses, and files in the executable and loaded libraries",
    "Use `command alias` to create custom shortcuts for frequently used commands",
    "LLDB's `register read` shows the contents of registers in the selected frame",
    "The `disassemble` command shows assembly instructions for the current function",
    "Use `thread step-inst` (or `si`) to step one instruction",
    "LLDB's Python API can be accessed with the `script` command to extend debugging capabilities",
    "Use `process attach --pid <pid>` to attach to a running process",
    "The `breakpoint command add` lets you run commands when a breakpoint is hit",
    "Use `memory find` to search for a value in the process's memory",
    "LLDB's `settings set` command allows you to customize debugger behavior",
    "The `platform list` command shows all available platforms for remote debugging",
]


def get_tip_of_the_day(debugger_type: int) -> str:
    """
    Returns a random tip based on the current debugger type.

    You should pass in pwndbg.dbg.name().value .
    """
    return choice(get_all_tips(debugger_type))


def get_all_tips(debugger_type: int) -> list[str]:
    """
    Returns all tips applicable to the current debugger.

    You should pass in pwndbg.dbg.name().value .
    """

    # Needs to mirror pwndbg/dbg_mod/__init__.py:DebuggerType
    class DebuggerType(Enum):
        GDB = 1
        LLDB = 2

    valid_values = [e.value for e in DebuggerType]
    assert debugger_type in valid_values

    match debugger_type:
        case DebuggerType.GDB.value:
            return GDB_TIPS + PWNDBG_TIPS
        case DebuggerType.LLDB.value:
            return LLDB_TIPS + PWNDBG_TIPS
        case _:
            return PWNDBG_TIPS


def color_tip(tip: str) -> str:
    return re.sub("`(.*?)`", lambda s: message.warn(s.group()[1:-1]), tip)



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class SigInfoKill:
    si_pid: int
    si_uid: int


@dataclass
class SigInfoTimer:
    si_tid: int
    si_overrun: int
    si_sigval: SigInfoSigVal


@dataclass
class SigInfoSigVal:
    sival_int: int
    sival_ptr: int


@dataclass
class SigInfoRt:
    si_pid: int
    si_uid: int
    si_sigval: SigInfoSigVal


@dataclass
class SigInfoSigChld:
    si_pid: int
    si_uid: int
    si_status: int
    si_utime: int
    si_stime: int


@dataclass
class SigInfoSigFault:
    si_addr: int


@dataclass
class SigInfoSigPoll:
    si_band: int
    si_fd: int


@dataclass
class SigInfoSigSys:
    call_addr: int
    syscall: int
    arch: int


@dataclass
class SigInfo:
    si_signo: int
    si_errno: int
    si_code: int

    kill: SigInfoKill
    timer: SigInfoTimer
    rt: SigInfoRt
    sigchld: SigInfoSigChld
    sigfault: SigInfoSigFault
    sigpoll: SigInfoSigPoll
    sigsys: SigInfoSigSys



"""
Perform queries specific to the musl libc.
"""

from __future__ import annotations

import pwndbg.aglib.elf
import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.lib.cache

from . import util
from .dispatch import LibcType
from .dispatch import LibcURLs


def type() -> LibcType:
    return LibcType.MUSL


@pwndbg.lib.cache.cache_until("start", "objfile")
def version(libc_filepath: str) -> tuple[int, ...]:
    # __libc_version is an internal symbol in musl, added in version v1.1.21
    # https://elixir.bootlin.com/musl/v1.1.21/source/src/internal/version.c#L4
    addr = pwndbg.aglib.symbol.lookup_symbol_addr("__libc_version", objfile_endswith=libc_filepath)
    if addr is not None:
        ver = pwndbg.aglib.memory.string(addr)
        return util.version_parse(ver)

    # The version string is simply not embedded into older versions of musl afaict.
    return (-1, -1)


@pwndbg.lib.cache.cache_until("start", "objfile")
def has_internal_symbols(libc_filepath: str) -> bool:
    # c_messages is an internal global variable in musl. Has existed since the
    # first release i.e. version v0.5.0 (2011). (elixir doesn't have 0.5.0 on hand)
    # https://elixir.bootlin.com/musl/v0.5.9/source/src/locale/langinfo.c#L24
    return (
        pwndbg.aglib.symbol.lookup_symbol("c_messages", objfile_endswith=libc_filepath) is not None
    )


@pwndbg.lib.cache.cache_until("start", "objfile")
def has_debug_info() -> bool:
    # Available since the first release (0.5.0). (elixir doesn't have 0.5.0 on hand)
    # https://elixir.bootlin.com/musl/v0.5.9/source/include/bits/pthread.h#L1
    return pwndbg.aglib.typeinfo.load("struct __ptcb") is not None


def verify_libc_candidate(mapping_name: str) -> bool:
    # First check __freadahead which is an exported symbol in musl, and bionic but not in glibc
    # It was introduced in v0.9.2 (year 2012)
    # https://elixir.bootlin.com/musl/v0.9.2/source/src/stdio/ext2.c#L3
    if (
        util.has_exported_symbols(mapping_name)
        and pwndbg.aglib.symbol.lookup_symbol("__freadahead", objfile_endswith=mapping_name) is None
    ):
        return False

    # Then do a consistent but more expensive (?) check:
    # Check if the string "/tmp/tmpnam_XXXX" is in the .rodata of the binary.
    # Added in musl version v1.1.2 (year 2014) (is present until at least v1.2.5).
    # https://elixir.bootlin.com/musl/v1.1.2/source/src/stdio/tmpnam.c#L15
    rodata: tuple[int, int, bytes] | None = pwndbg.aglib.elf.section_by_name(
        mapping_name, ".rodata", try_local_path=True
    )
    if rodata is None:
        return False
    _, _, data = rodata
    return b"/tmp/tmpnam_XXXX" in data


def verify_ld_candidate(mapping_name: str) -> bool:
    # For musl, ld and libc are the same mapping.
    # On some distributions it is named libc, on some it's ld.
    return verify_libc_candidate(mapping_name)


def urls(ver: tuple[int, ...] | None) -> LibcURLs:
    assert ver is not None
    if ver[0] == -1:
        # Version not available, use dummy values.
        return LibcURLs(
            versioned_readable_source="https://elixir.bootlin.com/musl/latest/source",
            versioned_compressed_source="https://musl.libc.org/releases/musl-<major>.<minor>.<patch>.tar.gz",
            homepage="https://musl.libc.org/",
            git="git://git.musl-libc.org/musl",
        )

    ver_str = ".".join(map(str, ver))
    return LibcURLs(
        versioned_readable_source=f"https://elixir.bootlin.com/musl/v{ver_str}/source",
        versioned_compressed_source=f"https://musl.libc.org/releases/musl-{ver_str}.tar.gz",
        homepage="https://musl.libc.org/",
        git="git://git.musl-libc.org/musl",
    )


def libc_same_as_ld() -> bool:
    return True



from __future__ import annotations

import typing
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from enum import auto
from typing import Literal

from capstone6pwndbg import CS_MODE_MICRO
from capstone6pwndbg import CS_MODE_MIPS1
from capstone6pwndbg import CS_MODE_MIPS2
from capstone6pwndbg import CS_MODE_MIPS3
from capstone6pwndbg import CS_MODE_MIPS4
from capstone6pwndbg import CS_MODE_MIPS5
from capstone6pwndbg import CS_MODE_MIPS32
from capstone6pwndbg import CS_MODE_MIPS32R2
from capstone6pwndbg import CS_MODE_MIPS32R3
from capstone6pwndbg import CS_MODE_MIPS32R5
from capstone6pwndbg import CS_MODE_MIPS32R6
from capstone6pwndbg import CS_MODE_MIPS64
from capstone6pwndbg import CS_MODE_MIPS64R2
from capstone6pwndbg import CS_MODE_MIPS64R3
from capstone6pwndbg import CS_MODE_MIPS64R5
from capstone6pwndbg import CS_MODE_MIPS64R6
from capstone6pwndbg import CS_MODE_NANOMIPS

# Names of arches that Pwndbg supports
# The names that GDB/LLDB recognize are converted to a name in this list
PWNDBG_SUPPORTED_ARCHITECTURES_TYPE = Literal[
    "x86-64",
    "i386",
    "i8086",
    "mips",
    "aarch64",
    "arm",
    "armcm",
    "rv32",
    "rv64",
    "sparc",
    "powerpc",
    "loongarch64",
    "s390x",
]

PWNDBG_SUPPORTED_ARCHITECTURES: list[PWNDBG_SUPPORTED_ARCHITECTURES_TYPE] = list(
    typing.get_args(PWNDBG_SUPPORTED_ARCHITECTURES_TYPE)
)

# mapping between pwndbg and pwntools arch names
PWNLIB_ARCH_MAPPINGS = {
    "x86-64": "amd64",
    "i386": "i386",
    "i8086": "none",
    "mips": "mips",
    "aarch64": "aarch64",
    "arm": "arm",
    "armcm": "thumb",
    "rv32": "riscv32",
    "rv64": "riscv64",
    "powerpc": "powerpc",
    "sparc": "sparc",
    "loongarch64": "none",
    "s390x": "s390",  # FIXME: I believe this should be s390x, but that's not supported
}


# The platform being debugged
class Platform(Enum):
    LINUX = (auto(),)
    DARWIN = (auto(),)
    # WINDOWS = auto(),
    # ANDROID = auto(),
    # OPENBSD = auto(),
    # FREEBSD = auto(),


PWNLIB_PLATFORM_MAPPINGS: dict[Platform, str] = {
    Platform.LINUX: "linux",
    Platform.DARWIN: "darwin",
}


# A series of tuples
# Index 0 = Unique integer
# Index 1 = Capstone constant associated with the attribute
class ArchAttribute(Enum):
    # MIPS ATTRIBUTES
    ## MIPS created and removed lots of instructions between ISA versions, sometimes re-using
    ## old instruction encodings. To disassemble correctly, we need to choose the correct ISA.
    MIPS_ISA_1 = auto(), CS_MODE_MIPS1
    MIPS_ISA_2 = auto(), CS_MODE_MIPS2
    MIPS_ISA_3 = auto(), CS_MODE_MIPS3
    MIPS_ISA_4 = auto(), CS_MODE_MIPS4
    MIPS_ISA_5 = auto(), CS_MODE_MIPS5

    MIPS_ISA_32 = auto(), CS_MODE_MIPS32
    MIPS_ISA_32R2 = auto(), CS_MODE_MIPS32R2

    MIPS_ISA_32R3 = auto(), CS_MODE_MIPS32R3
    MIPS_ISA_32R5 = auto(), CS_MODE_MIPS32R5
    MIPS_ISA_32R6 = auto(), CS_MODE_MIPS32R6

    MIPS_ISA_64 = auto(), CS_MODE_MIPS64
    MIPS_ISA_64R2 = auto(), CS_MODE_MIPS64R2
    MIPS_ISA_64R3 = auto(), CS_MODE_MIPS64R3
    MIPS_ISA_64R5 = auto(), CS_MODE_MIPS64R5
    MIPS_ISA_64R6 = auto(), CS_MODE_MIPS64R6

    MIPS_ISA_MICRO = auto(), CS_MODE_MICRO
    MIPS_ISA_NANO = auto(), CS_MODE_NANOMIPS

    def __init__(self, _, cs_mode):
        self.cs_mode = cs_mode


@dataclass
class ArchDefinition:
    name: PWNDBG_SUPPORTED_ARCHITECTURES_TYPE
    ptrsize: int
    """Pointer size in bytes"""
    endian: Literal["little", "big"]
    platform: Platform
    attributes: list[ArchAttribute] = field(default_factory=list)



from __future__ import annotations

import argparse

import pwndbg.aglib
import pwndbg.aglib.proc
import pwndbg.commands
from pwndbg.color import context
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Print out ARM CPSR or xPSR register.")

parser.add_argument(
    "cpsr_value", help="Parse the given CPSR value instead of the actual one.", nargs="?", type=int
)


@pwndbg.commands.Command(
    parser,
    aliases=["xpsr", "pstate"],
    category=CommandCategory.REGISTER,
)
@pwndbg.aglib.proc.OnlyWithArch(["arm", "armcm", "aarch64"])
@pwndbg.commands.OnlyWhenRunning
def cpsr(cpsr_value=None) -> None:
    reg = "xpsr" if pwndbg.aglib.arch.name == "armcm" else "cpsr"
    reg_flags = pwndbg.aglib.regs.flags[reg]

    if cpsr_value is not None:
        reg_val = cpsr_value
    else:
        reg_val = pwndbg.aglib.regs.read_reg(reg)

    print(f"{reg} {context.format_flags(reg_val, reg_flags)}")



"""
Compatibility functionality for WinDbg users.
"""

from __future__ import annotations

import argparse
import codecs
from itertools import chain

import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.strings
import pwndbg.aglib.symbol
import pwndbg.commands
import pwndbg.commands.hexdump
import pwndbg.commands.next
from pwndbg.commands import CommandCategory

if pwndbg.dbg.is_gdblib_available():
    import gdb


def enhex(size, value):
    value = value & ((1 << 8 * size) - 1)
    x = f"{abs(value):x}"
    x = x.rjust(size * 2, "0")
    return x


# `pwndbg.hexdump` imports `enhex` from this module, so we have to import it
# after it's been defined in order to avoid circular import errors.
import pwndbg.hexdump

parser = argparse.ArgumentParser(description="Starting at the specified address, dump N bytes.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to dump from."
)
parser.add_argument(
    "count",
    type=pwndbg.commands.AddressExpr,
    default=64,
    nargs="?",
    help="The number of bytes to dump.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def db(address, count=64):
    """
    Starting at the specified address, dump N bytes
    (default 64).
    """
    return dX(1, address, count, repeat=db.repeat)


parser = argparse.ArgumentParser(description="Starting at the specified address, dump N words.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to dump from."
)
parser.add_argument(
    "count",
    type=pwndbg.commands.AddressExpr,
    default=32,
    nargs="?",
    help="The number of words to dump.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def dw(address, count=32):
    """
    Starting at the specified address, dump N words
    (default 32).
    """
    return dX(2, address, count, repeat=dw.repeat)


parser = argparse.ArgumentParser(description="Starting at the specified address, dump N dwords.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to dump from."
)
parser.add_argument(
    "count",
    type=pwndbg.commands.AddressExpr,
    default=16,
    nargs="?",
    help="The number of dwords to dump.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def dd(address, count=16):
    """
    Starting at the specified address, dump N dwords
    (default 16).
    """
    return dX(4, address, count, repeat=dd.repeat)


parser = argparse.ArgumentParser(description="Starting at the specified address, dump N qwords.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to dump from."
)
parser.add_argument(
    "count",
    type=pwndbg.commands.AddressExpr,
    default=8,
    nargs="?",
    help="The number of qwords to dump.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def dq(address, count=8):
    """
    Starting at the specified address, dump N qwords
    (default 8).
    """
    return dX(8, address, count, repeat=dq.repeat)


parser = argparse.ArgumentParser(description="Starting at the specified address, hexdump.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to dump from."
)
parser.add_argument(
    "count",
    type=pwndbg.commands.AddressExpr,
    default=8,
    nargs="?",
    help="The number of bytes to hexdump.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def dc(address, count=8):
    return pwndbg.commands.hexdump.hexdump(address=address, count=count)


def dX(size, address, count, to_string=False, repeat=False):
    """
    Traditionally, windbg will display 16 bytes of data per line.
    """

    lines = list(
        chain.from_iterable(
            pwndbg.hexdump.hexdump(
                data=None, size=size, count=count, address=address, repeat=repeat, dX_call=True
            )
        )
    )

    if not to_string and lines:
        print("\n".join(lines))

    return lines


parser = argparse.ArgumentParser(description="Write hex bytes at the specified address.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to write to."
)
parser.add_argument("data", type=str, nargs="*", help="The bytes to write.")


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def eb(address, data):
    """
    Write hex bytes at the specified address.
    """
    return eX(1, address, data)


parser = argparse.ArgumentParser(description="Write hex words at the specified address.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to write to."
)
parser.add_argument("data", type=str, nargs="*", help="The words to write.")


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def ew(address, data):
    """
    Write hex words at the specified address.
    """
    return eX(2, address, data)


parser = argparse.ArgumentParser(description="Write hex dwords at the specified address.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to write to."
)
parser.add_argument("data", type=str, nargs="*", help="The dwords to write.")


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def ed(address, data):
    """
    Write hex dwords at the specified address.
    """
    return eX(4, address, data)


parser = argparse.ArgumentParser(description="Write hex qwords at the specified address.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to write to."
)
parser.add_argument("data", type=str, nargs="*", help="The qwords to write.")


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def eq(address, data):
    """
    Write hex qwords at the specified address.
    """
    return eX(8, address, data)


parser = argparse.ArgumentParser(description="Write a string at the specified address.")
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to write to."
)
parser.add_argument("data", type=str, help="The string to write.")


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def ez(address, data):
    """
    Write a character at the specified address.
    """
    return eX(1, address, data, hex=False)


parser = argparse.ArgumentParser(
    description="Write a string at the specified address."
)  # TODO Is eza just ez? If so just alias. I had trouble finding windbg documentation defining ez
parser.add_argument(
    "address", type=pwndbg.commands.HexOrAddressExpr, help="The address to write to."
)
parser.add_argument("data", type=str, help="The string to write.")


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def eza(address, data):
    """
    Write a string at the specified address.
    """
    return ez(address, data)


def eX(size, address, data, hex=True) -> None:
    """
    This relies on windbg's default hex encoding being enforced
    """
    if not data:
        print("Cannot write empty data into memory.")
        return

    if hex:
        # Early validation if all data is hex
        for string in data:
            if string.startswith("0x"):
                string = string[2:]

            if any(ch not in "0123456789abcdefABCDEF" for ch in string):
                print(
                    "Incorrect data format: it must all be a hex value (0x1234 or 1234, both interpreted as 0x1234)"
                )
                return

    writes = 0
    for i, string in enumerate(data):
        if hex:
            if string.startswith("0x"):
                string = string[2:]

            string = string.rjust(size * 2, "0")

            data = codecs.decode(string, "hex")
        else:
            data = string

        if pwndbg.aglib.arch.endian == "little":
            data = data[::-1]

        try:
            pwndbg.aglib.memory.write(address + (i * size), data)
            writes += 1
        except pwndbg.dbg_mod.Error:
            print(f"Cannot access memory at address {address:#x}")
            if writes > 0:
                print(f"(Made {writes} writes to memory; skipping further writes)")
            return


parser = argparse.ArgumentParser(description="Dump pointers and symbols at the specified address.")
parser.add_argument("addr", type=pwndbg.commands.HexOrAddressExpr, help="The address to dump from.")


@pwndbg.commands.Command(
    parser, aliases=["kd", "dps", "dqs"], category=CommandCategory.WINDBG
)  # TODO are these really all the same? They had identical implementation...
@pwndbg.commands.OnlyWhenRunning
def dds(addr):
    """
    Dump pointers and symbols at the specified address.
    """
    return pwndbg.commands.telescope.telescope(addr)


da_parser = argparse.ArgumentParser(description="Dump a string at the specified address.")
da_parser.add_argument("address", type=pwndbg.commands.HexOrAddressExpr, help="Address to dump")
da_parser.add_argument("max", type=int, nargs="?", default=256, help="Maximum string length")


@pwndbg.commands.Command(da_parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def da(address, max) -> None:
    print(f"{address:x}", repr(pwndbg.aglib.strings.get(address, max)))


ds_parser = argparse.ArgumentParser(description="Dump a string at the specified address.")
ds_parser.add_argument("address", type=pwndbg.commands.HexOrAddressExpr, help="Address to dump")
ds_parser.add_argument("max", type=int, nargs="?", default=256, help="Maximum string length")


@pwndbg.commands.Command(ds_parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def ds(address, max) -> None:
    # We do change the max length to the default if its too low
    # because the truncated display is not that ideal/not the same as GDB's yet
    # (ours: "truncated ...", GDBs: "truncated "...)
    if max < 256:
        print(f"Max str len of {max} too low, changing to 256")
        max = 256

    string = pwndbg.aglib.strings.get(address, max, maxread=4096)
    if string:
        print(f"{address:x} {string!r}")
    else:
        print(
            "Data at address can't be dereferenced or is not a printable null-terminated string or is too short."
        )
        print("Perhaps try: db <address> <count> or hexdump <address>")


if pwndbg.dbg.is_gdblib_available():

    @pwndbg.commands.Command("List breakpoints.", category=CommandCategory.WINDBG)
    def bl() -> None:
        """
        List breakpoints
        """
        gdb.execute("info breakpoints")

    parser = argparse.ArgumentParser(description="Disable the breakpoint with the specified index.")
    parser.add_argument(
        "which", nargs="?", type=str, default="*", help="Index of the breakpoint to disable."
    )

    @pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
    def bd(which="*") -> None:
        """
        Disable the breakpoint with the specified index.
        """
        if which == "*":
            gdb.execute("disable breakpoints")
        else:
            gdb.execute(f"disable breakpoints {which}")

    parser = argparse.ArgumentParser(description="Enable the breakpoint with the specified index.")
    parser.add_argument(
        "which", nargs="?", type=str, default="*", help="Index of the breakpoint to enable."
    )

    @pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
    def be(which="*") -> None:
        """
        Enable the breakpoint with the specified index.
        """
        if which == "*":
            gdb.execute("enable breakpoints")
        else:
            gdb.execute(f"enable breakpoints {which}")

    parser = argparse.ArgumentParser(description="Clear the breakpoint with the specified index.")
    parser.add_argument(
        "which", nargs="?", type=str, default="*", help="Index of the breakpoint to clear."
    )

    @pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
    def bc(which="*") -> None:
        """
        Clear the breakpoint with the specified index.
        """
        if which == "*":
            gdb.execute("delete breakpoints")
        else:
            gdb.execute(f"delete breakpoints {which}")

    parser = argparse.ArgumentParser(description="Set a breakpoint at the specified address.")
    parser.add_argument("where", type=int, help="The address to break at.")

    @pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
    def bp(where) -> None:
        gdb.execute(f"break *{where:#x}")

    @pwndbg.commands.Command("Print a backtrace (alias 'bt').", category=CommandCategory.WINDBG)
    @pwndbg.commands.OnlyWhenRunning
    def k() -> None:
        """
        Print a backtrace (alias 'bt')
        """
        gdb.execute("bt")

    @pwndbg.commands.Command(
        "WinDbg compatibility alias for 'continue' command.", category=CommandCategory.WINDBG
    )
    @pwndbg.commands.OnlyWhenRunning
    def go() -> None:
        """
        WinDbg compatibility alias for 'continue' command.
        """
        gdb.execute("continue")


parser = argparse.ArgumentParser(description="List the symbols nearest to the provided value.")
parser.add_argument(
    "value", type=int, nargs="?", default=None, help="The address you want the name of."
)


@pwndbg.commands.Command(parser, category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def ln(value: int = None) -> None:
    """
    List the symbols nearest to the provided value.
    """
    if value is None:
        value = pwndbg.aglib.regs.pc

    x = pwndbg.aglib.symbol.resolve_addr(value)
    if x:
        result = f"({value:#x})   {x}"
        print(result)


# The three commands are aliases for `vmmap` and are set so in vmmap.py
# lm
# address
# vprot


@pwndbg.commands.Command("Not be windows.", category=CommandCategory.WINDBG)
@pwndbg.commands.OnlyWhenRunning
def peb() -> None:
    print("This isn't Windows!")


@pwndbg.commands.Command(
    "WinDbg compatibility alias for 'nextcall' command.", category=CommandCategory.WINDBG
)
@pwndbg.commands.OnlyWhenRunning
def pc():
    """
    WinDbg compatibility alias for 'nextcall' command.
    """
    return pwndbg.commands.next.nextcall()



"""
Library for handling GDB logic. Being removed in favour of aglib.
"""

from __future__ import annotations

import re

import gdb

from pwndbg.gdblib import config as config_mod

regs = None

__all__ = ()

# Export parsed GDB version
gdb_version = tuple(map(int, re.search(r"(\d+)[^\d]+(\d+)", gdb.VERSION).groups()))
if gdb_version[0] < 12:
    msg = "Unsupported GDB version, pwndbg only support GDB12+"
    print(msg)
    raise RuntimeError(msg)


# TODO: should the imports above be moved here?
def load_gdblib() -> None:
    """
    Import all gdblib modules that need to run code on import
    """
    # pylint: disable=import-outside-toplevel
    import pwndbg.gdblib.bpoint
    import pwndbg.gdblib.functions
    import pwndbg.gdblib.got
    import pwndbg.gdblib.prompt
    import pwndbg.gdblib.symbol
    import pwndbg.gdblib.tui
